#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.xcube.rules.increasing_time import IncreasingTime
from xrlint.testing import RuleTest, RuleTester


def make_dataset():
    dims = ["time", "y", "x"]
    n = 5
    return xr.Dataset(
        attrs=dict(title="v-data"),
        coords={
            dims[2]: xr.DataArray(
                np.linspace(0, 1, n), dims=dims[2], attrs={"units": "m"}
            ),
            dims[1]: xr.DataArray(
                np.linspace(0, 1, n), dims=dims[1], attrs={"units": "m"}
            ),
            dims[0]: xr.DataArray(
                [2010, 2011, 2012, 2013, 2014],
                dims=dims[0],
                attrs={"units": "years"},
            ),
        },
        data_vars={
            "chl": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
            "tsm": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
            "mask": xr.DataArray(np.random.random((n, n)), dims=dims[-2:]),
        },
    )


valid_dataset_1 = make_dataset()

invalid_dataset_1 = make_dataset()
invalid_dataset_1 = invalid_dataset_1.assign_coords(
    time=xr.DataArray(
        [2010, 2011, 2012, 2013, 2013], dims="time", attrs={"units": "years"}
    )
)

invalid_dataset_2 = make_dataset()
invalid_dataset_2 = invalid_dataset_2.assign_coords(
    time=xr.DataArray(
        [2010, 2011, 2012, 2014, 2013], dims="time", attrs={"units": "years"}
    )
)

LatLonNamingTest = RuleTester.define_test(
    "increasing-time",
    IncreasingTime,
    valid=[
        RuleTest(dataset=valid_dataset_1),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_1, expected=1),
        RuleTest(dataset=invalid_dataset_2, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from xrlint.plugins.xcube.rules.dataset_title import DatasetTitle
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset(attrs=dict(title="OC-Climatology"))
valid_dataset_1 = xr.Dataset(attrs=dict(title="SST-Climatology"))

invalid_dataset_0 = xr.Dataset()
invalid_dataset_1 = xr.Dataset(attrs=dict(title=""))


DatasetTitleTest = RuleTester.define_test(
    "dataset-title",
    DatasetTitle,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=1),
        RuleTest(dataset=invalid_dataset_1, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.util.naming import to_kebab_case, to_snake_case


class NamingTest(TestCase):
    def test_to_kebab_case(self):
        self.assertEqual("", to_kebab_case(""))
        self.assertEqual("rule", to_kebab_case("rule"))
        self.assertEqual("rule", to_kebab_case("Rule"))
        self.assertEqual("my-rule", to_kebab_case("MyRule"))
        self.assertEqual("my-rule-a", to_kebab_case("MyRuleA"))
        self.assertEqual("my-rule-3", to_kebab_case("MyRule3"))
        self.assertEqual("my-rule-3", to_kebab_case("MyRule_3"))
        self.assertEqual("my-rule-3", to_kebab_case("My_Rule_3"))
        self.assertEqual("my-rule-3", to_kebab_case("my-rule-3"))
        self.assertEqual("my-rule-3", to_kebab_case("My-Rule-3"))
        self.assertEqual("my-rule-3", to_kebab_case("My-Rule 3"))
        self.assertEqual("abc-rule-123", to_kebab_case("ABCRule123"))
        self.assertEqual("abc-rule-xyz", to_kebab_case("ABCRuleXYZ"))

    def test_to_snake_case(self):
        self.assertEqual("", to_snake_case(""))
        self.assertEqual("rule", to_snake_case("rule"))
        self.assertEqual("rule", to_snake_case("Rule"))
        self.assertEqual("my_rule", to_snake_case("MyRule"))
        self.assertEqual("my_rule_a", to_snake_case("MyRuleA"))
        self.assertEqual("my_rule_3", to_snake_case("MyRule3"))
        self.assertEqual("my_rule_3", to_snake_case("MyRule_3"))
        self.assertEqual("my_rule_3", to_snake_case("My_Rule_3"))
        self.assertEqual("my_rule_3", to_snake_case("my-rule-3"))
        self.assertEqual("my_rule_3", to_snake_case("My-Rule-3"))
        self.assertEqual("my_rule_3", to_snake_case("My-Rule 3"))
        self.assertEqual("abc_rule_123", to_snake_case("ABCRule123"))
        self.assertEqual("abc_rule_xyz", to_snake_case("ABCRuleXYZ"))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import math

from xrlint.node import VariableNode
from xrlint.plugins.xcube.plugin import plugin
from xrlint.rule import RuleContext, RuleOp
from xrlint.util.schema import schema

DEFAULT_LIMIT = 5


@plugin.define_rule(
    "no-chunked-coords",
    version="1.0.0",
    type="problem",
    description=(
        "Coordinate variables should not be chunked."
        " Can be used to identify performance issues, where chunked coordinates"
        " can cause slow opening if datasets due to the many chunk-fetching"
        " requests made to (remote) filesystems with low bandwidth."
        " You can use the `limit` parameter to specify an acceptable number "
        f" of chunks. Its default is {DEFAULT_LIMIT}."
    ),
    schema=schema(
        "object",
        properties=dict(
            limit=schema(
                "integer",
                minimum=0,
                default=DEFAULT_LIMIT,
                title="Acceptable number of chunks",
            )
        ),
    ),
)
class NoChunkedCoords(RuleOp):
    def __init__(self, limit: int = DEFAULT_LIMIT):
        self.limit = limit

    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        if node.name not in ctx.dataset.coords or node.array.ndim != 1:
            return

        chunks = node.array.encoding.get("chunks")
        if isinstance(chunks, (list, tuple)) and len(chunks) == 1:
            num_chunks = math.ceil(node.array.size / chunks[0])
            if num_chunks > self.limit:
                ctx.report(
                    f"Number of chunks exceeds limit: {num_chunks} > {self.limit}.",
                    suggestions=["Combine chunks into a one or more larger ones."],
                )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from dataclasses import dataclass, field
from typing import Any, Callable, Literal, Type

from xrlint.config import Config, ConfigLike, ConfigObject
from xrlint.processor import Processor, ProcessorOp, define_processor
from xrlint.rule import Rule, RuleOp, define_rule
from xrlint.util.constructible import MappingConstructible
from xrlint.util.importutil import import_value
from xrlint.util.serializable import JsonSerializable, JsonValue


@dataclass(kw_only=True)
class PluginMeta(MappingConstructible, JsonSerializable):
    """Plugin metadata."""

    name: str
    """Plugin name."""

    version: str = "0.0.0"
    """Plugin version."""

    ref: str | None = None
    """Plugin reference.
    Specifies the location from where the plugin can be
    dynamically imported.
    Must have the form "<module>:<attr>", if given.
    """

    docs_url: str | None = None
    """Plugin documentation URL."""

    @classmethod
    def value_name(cls) -> str:
        return "plugin_meta"

    @classmethod
    def value_type_name(cls) -> str:
        return "PluginMeta | dict"


@dataclass(frozen=True, kw_only=True)
class Plugin(MappingConstructible, JsonSerializable):
    """A plugin that can contribute rules, processors,
    and predefined configurations to XRLint.

    Use the factory [new_plugin][xrlint.plugin.new_plugin]
    to create plugin instances.
    """

    meta: PluginMeta
    """Information about the plugin."""

    rules: dict[str, Rule] = field(default_factory=dict)
    """A dictionary containing the definitions of custom rules."""

    processors: dict[str, Processor] = field(default_factory=dict)
    """A dictionary containing named processors.
    """

    configs: dict[str, list[ConfigObject]] = field(default_factory=dict)
    """A dictionary containing named configuration lists."""

    def define_rule(
        self,
        name: str,
        version: str = "0.0.0",
        schema: dict[str, Any] | list[dict[str, Any]] | bool | None = None,
        type: Literal["problem", "suggestion", "layout"] = "problem",
        description: str | None = None,
        docs_url: str | None = None,
        op_class: Type[RuleOp] | None = None,
    ) -> Callable[[Any], Type[RuleOp]] | None:
        """Decorator to define a plugin rule.
        The method registers a new rule with the plugin.

        Refer to [define_rule][xrlint.rule.define_rule]
        for details.
        """
        return define_rule(
            name=name,
            version=version,
            schema=schema,
            type=type,
            description=description,
            docs_url=docs_url,
            op_class=op_class,
            registry=self.rules,
        )

    def define_processor(
        self,
        name: str | None = None,
        version: str = "0.0.0",
        op_class: Type[ProcessorOp] | None = None,
    ):
        """Decorator to define a plugin processor.
        The method registers a new processor with the plugin.

        Refer to [define_processor][xrlint.processor.define_processor]
        for details.
        """
        return define_processor(
            name=name,
            version=version,
            op_class=op_class,
            registry=self.processors,
        )

    def define_config(self, name: str, config: ConfigLike) -> Config:
        """Define a named configuration.

        Args:
            name: The name of the configuration.
            config: A configuration-like value.
                For more information see the
                [ConfigLike][xrlint.config.ConfigLike] type alias.

        Returns:
            The configuration.
        """
        config = Config.from_value(config)
        self.configs[name] = list(config.objects)
        return config

    @classmethod
    def _from_str(cls, value: str, value_name: str) -> "Plugin":
        plugin, plugin_ref = import_value(
            value, "export_plugin", factory=Plugin.from_value
        )
        plugin.meta.ref = plugin_ref
        return plugin

    @classmethod
    def value_name(cls) -> str:
        return "plugin"

    @classmethod
    def value_type_name(cls) -> str:
        return "Plugin | dict | str"

    def to_json(self, value_name: str | None = None) -> JsonValue:
        if self.meta.ref:
            return self.meta.ref
        return super().to_json(value_name=value_name)


def new_plugin(
    name: str,
    version: str = "0.0.0",
    ref: str | None = None,
    docs_url: str | None = None,
    rules: dict[str, Rule] | None = None,
    processors: dict[str, Processor] | None = None,
    configs: dict[str, ConfigObject] | None = None,
) -> Plugin:
    """Create a new plugin object that can contribute rules, processors,
    and predefined configurations to XRLint.

    Args:
        name: Plugin name. Required.
        version: Plugin version. Defaults to `"0.0.0"`.
        ref: Plugin reference. Optional.
        docs_url: Plugin documentation URL. Optional.
        rules: A dictionary containing the definitions of custom rules. Optional.
        processors: A dictionary containing custom processors. Optional.
        configs: A dictionary containing predefined configurations. Optional.
    """
    return Plugin(
        meta=PluginMeta(name=name, version=version, ref=ref, docs_url=docs_url),
        rules=rules or {},
        processors=processors or {},
        configs=configs or {},
    )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import unittest
from unittest import TestCase

import pytest

from xrlint.rule import Rule, RuleConfig, RuleMeta, RuleOp, define_rule


class MyRule1(RuleOp):
    """This is my 1st rule."""


class MyRule2(RuleOp):
    """This is my 2nd rule."""


def export_rule():
    return Rule(meta=RuleMeta(name="my-rule-1"), op_class=MyRule1)


class RuleTest(TestCase):
    def test_class_props(self):
        self.assertIs(RuleMeta, Rule.meta_class())
        self.assertIs(RuleOp, Rule.op_base_class())
        self.assertEqual("rule", Rule.value_name())
        self.assertEqual("Rule | Type[RuleOp] | dict | str", Rule.value_type_name())

    def test_from_value_ok_rule(self):
        rule = export_rule()
        rule2 = Rule.from_value(rule)
        self.assertIs(rule, rule2)

    def test_from_value_ok_rule_op(self):
        rule = export_rule()
        rule2 = Rule.from_value(rule)
        self.assertIs(rule, rule2)

    def test_from_value_ok_str(self):
        rule = Rule.from_value("tests.test_rule")
        self.assertIsInstance(rule, Rule)
        self.assertEqual("my-rule-1", rule.meta.name)
        self.assertIs(MyRule1, rule.op_class)

    # noinspection PyMethodMayBeStatic
    def test_from_value_fails(self):
        with pytest.raises(
            TypeError, match="value must be of type Rule|str, but got int"
        ):
            Rule.from_value(73)

        class MyRule3(RuleOp):
            """This is my 3rd rule."""

        with pytest.raises(
            ValueError,
            match=r"missing rule metadata, apply define_rule\(\) to class MyRule3",
        ):
            Rule.from_value(MyRule3)

    def test_to_json(self):
        class MyRule3(RuleOp):
            """This is my 3rd rule."""

        rule = Rule.from_value("tests.test_rule")
        self.assertEqual("tests.test_rule:export_rule", rule.to_json())

        rule = Rule(meta=RuleMeta(name="r3", ref="mymod.rules:r3"), op_class=MyRule3)
        self.assertEqual("mymod.rules:r3", rule.to_json())

        rule = Rule(meta=RuleMeta(name="r3"), op_class=MyRule3)
        self.assertEqual(
            {
                "meta": {"name": "r3"},
                "op_class": "<class 'tests.test_rule.RuleTest.test_to_json.<locals>.MyRule3'>",
            },
            rule.to_json(),
        )


class RuleMetaTest(unittest.TestCase):
    def test_class_props(self):
        self.assertEqual("rule_meta", RuleMeta.value_name())
        self.assertEqual("RuleMeta | dict", RuleMeta.value_type_name())

    def test_from_value(self):
        rule_meta = RuleMeta.from_value(
            {
                "name": "r-4",
                "version": "2.0.1",
                "type": "suggestion",
                "description": "Be nice, always.",
            }
        )
        self.assertEqual(
            RuleMeta(
                name="r-4",
                version="2.0.1",
                type="suggestion",
                description="Be nice, always.",
            ),
            rule_meta,
        )

    def test_to_json(self):
        rule_meta = RuleMeta(name="r1", version="0.1.2", description="Nice one.")
        self.assertEqual(
            {"description": "Nice one.", "name": "r1", "version": "0.1.2"},
            rule_meta.to_json(),
        )


class DefineRuleTest(unittest.TestCase):
    def test_decorator(self):
        deco = define_rule()
        self.assertTrue(callable(deco))
        op_class = deco(MyRule1)
        self.assertIs(MyRule1, op_class)
        self.assertTrue(hasattr(MyRule1, "meta"))
        # noinspection PyUnresolvedReferences
        self.assertEqual("my-rule-1", MyRule1.meta.name)

    def test_function(self):
        rule = define_rule(op_class=MyRule1)
        self.assertIsInstance(rule, Rule)
        self.assertEqual("my-rule-1", rule.meta.name)
        self.assertIs(MyRule1, rule.op_class)

    def test_with_registry(self):
        registry = {}
        rule1 = define_rule(op_class=MyRule1, registry=registry)
        rule2 = define_rule(op_class=MyRule2, registry=registry)
        self.assertIs(rule1, registry["my-rule-1"])
        self.assertIs(rule2, registry["my-rule-2"])

    # noinspection PyMethodMayBeStatic
    def test_fail(self):
        with pytest.raises(
            TypeError,
            match="op_class must be a subclass of RuleOp, but got DefineRuleTest",
        ):
            # noinspection PyTypeChecker
            define_rule(op_class=DefineRuleTest)


class RuleConfigTest(TestCase):
    def test_class_props(self):
        self.assertEqual("rule_config", RuleConfig.value_name())
        self.assertEqual("int | str | list", RuleConfig.value_type_name())

    def test_defaults(self):
        rule_config = RuleConfig(1)
        self.assertEqual(1, rule_config.severity)
        self.assertEqual((), rule_config.args)
        self.assertEqual({}, rule_config.kwargs)

    def test_from_value_ok(self):
        self.assertEqual(RuleConfig(0), RuleConfig.from_value(0))
        self.assertEqual(RuleConfig(1), RuleConfig.from_value(1))
        self.assertEqual(RuleConfig(2), RuleConfig.from_value(2))
        self.assertEqual(RuleConfig(0), RuleConfig.from_value("off"))
        self.assertEqual(RuleConfig(1), RuleConfig.from_value("warn"))
        self.assertEqual(RuleConfig(2), RuleConfig.from_value("error"))
        self.assertEqual(RuleConfig(2), RuleConfig.from_value(["error"]))
        # YAML "on"/"off" literals
        self.assertEqual(RuleConfig(0), RuleConfig.from_value(False))
        self.assertEqual(RuleConfig(1), RuleConfig.from_value(True))

        self.assertEqual(
            RuleConfig(1, ("never",)), RuleConfig.from_value(["warn", "never"])
        )
        self.assertEqual(
            RuleConfig(1, ("always",)), RuleConfig.from_value([1, "always"])
        )
        self.assertEqual(
            RuleConfig(1, ("always", False)),
            RuleConfig.from_value([1, "always", False]),
        )
        self.assertEqual(
            RuleConfig(1, (), {"pattern": "*/*"}),
            RuleConfig.from_value([1, {"pattern": "*/*"}]),
        )
        self.assertEqual(
            RuleConfig(1, ("always",), {"pattern": "*/*"}),
            RuleConfig.from_value([1, "always", {"pattern": "*/*"}]),
        )
        self.assertEqual(
            RuleConfig(2, ("always", False), {"pattern": "*/*"}),
            RuleConfig.from_value([2, "always", False, {"pattern": "*/*"}]),
        )
        self.assertEqual(
            RuleConfig(0, ("always", {}), {"pattern": "*/*"}),
            RuleConfig.from_value(("off", "always", {}, {"pattern": "*/*"})),
        )

    # noinspection PyMethodMayBeStatic
    def test_from_value_fail(self):
        with pytest.raises(
            TypeError,
            match=r"rule_config must be of type int \| str \| list, but got None",
        ):
            RuleConfig.from_value(None)

        with pytest.raises(
            ValueError,
            match="severity must be one of 'error', 'warn', 'off', 2, 1, 0, but got 4",
        ):
            RuleConfig.from_value(4)

        with pytest.raises(
            ValueError,
            match=(
                "severity must be one of 'error', 'warn', 'off',"
                " 2, 1, 0, but got 'debug'"
            ),
        ):
            RuleConfig.from_value("debug")

        with pytest.raises(
            ValueError,
            match="rule_config must not be empty",
        ):
            RuleConfig.from_value([])



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from collections.abc import Hashable
from typing import Any

import xarray as xr

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.constants import TIME_NAME
from xrlint.plugins.xcube.plugin import plugin
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "time-naming",
    version="1.0.0",
    type="problem",
    description=f"Time coordinate and dimension should be called {TIME_NAME!r}.",
    docs_url="https://xcube.readthedocs.io/en/latest/cubespec.html#temporal-reference",
)
class TimeNaming(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        time_vars = {
            var_name: var
            for var_name, var in node.dataset.coords.items()
            if var_name != TIME_NAME and _is_time_coord(var_name, var)
        }
        for var_name, var in time_vars.items():
            ctx.report(
                f"The coordinate {var_name!r} should be named {TIME_NAME!r}.",
                suggestions=[f"Rename {var_name!r} to {TIME_NAME!r}."],
            )

        time_var = node.dataset.coords.get(TIME_NAME)
        if time_var is not None:
            if not _is_time_coord(TIME_NAME, time_var):
                ctx.report(f"Missing time units for coordinate {TIME_NAME!r}.")
            if not _get_time_encoding_attr(time_var, "calendar"):
                ctx.report(f"Missing calendar for coordinate {TIME_NAME!r}.")


def _is_time_coord(var_name: Hashable, var: xr.DataArray) -> bool:
    if var.dims == (var_name,):
        units = _get_time_encoding_attr(var, "units")
        return units and " since " in units
    return False


def _get_time_encoding_attr(var: xr.DataArray, name: str) -> Any:
    # decode_cf=True / decode_cf=False
    return var.encoding.get(name, var.attrs.get(name))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import VariableNode
from xrlint.plugins.xcube.plugin import plugin
from xrlint.plugins.xcube.util import is_spatial_var
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "data-var-colors",
    version="1.0.0",
    type="suggestion",
    description=(
        "Spatial data variables should encode xcube color mappings in their metadata."
    ),
    docs_url=(
        "https://xcube.readthedocs.io/en/latest/cubespec.html#encoding-of-colors"
    ),
)
class DataVarColors(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        array = node.array
        if not node.in_data_vars() or not is_spatial_var(array):
            return
        attrs = array.attrs
        color_bar_name = attrs.get("color_bar_name")
        if not color_bar_name:
            ctx.report("Missing attribute 'color_bar_name'.")
        else:
            color_value_min = attrs.get("color_value_min")
            color_value_max = attrs.get("color_value_max")
            if color_value_min is None or color_value_max is None:
                ctx.report(
                    "Missing both or one of 'color_value_min' and 'color_value_max'."
                )

            color_norm = attrs.get("color_norm")
            if color_norm and color_norm not in ("lin", "log"):
                ctx.report(
                    "Invalid value of attribute 'color_norm', should be 'lin' or 'log'."
                )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp
from xrlint.util.formatting import format_item


@plugin.define_rule(
    "coords-for-dims",
    version="1.0.0",
    type="problem",
    description="Dimensions of data variables should have corresponding coordinates.",
)
class CoordsForDims(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        dataset = node.dataset

        # Get data variable dimensions
        data_var_dims = set()
        for v in dataset.data_vars.values():
            data_var_dims.update(v.dims)
        if not data_var_dims:
            return

        # Get dimensions with coordinate variables
        no_coord_dims = []
        for d in sorted(str(d) for d in data_var_dims):
            if d not in dataset.coords:
                no_coord_dims.append(d)

        if no_coord_dims:
            n = len(no_coord_dims)
            ctx.report(
                f"{format_item(n, 'Data variable dimension')} without"
                f" coordinates: {', '.join(no_coord_dims)}.",
                suggestions=[
                    f"Add corresponding {format_item(n, 'coordinate variable')}"
                    f" to dataset:"
                    f" {', '.join(f'{d}[{dataset.sizes[d]}]' for d in no_coord_dims)}."
                ],
            )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from tests.plugins.xcube.helpers import make_cube
from xrlint.plugins.xcube.rules.no_chunked_coords import NoChunkedCoords
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset(attrs=dict(title="Empty"))
valid_dataset_1 = make_cube(360, 180, 3)
valid_dataset_2 = make_cube(90, 45, 20)
# ok, below default limit 5: ceil(20 / 5) = 4
valid_dataset_2.time.encoding["chunks"] = [4]

invalid_dataset_0 = make_cube(90, 45, 10)
# exceed default limit 5: ceil(10 / 1) = 10
invalid_dataset_0.time.encoding["chunks"] = [1]

NoChunkedCoordsTest = RuleTester.define_test(
    "no-chunked-coords",
    NoChunkedCoords,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Final


_MODULE_BASENAME: Final = "xrlint_config"
_REGULAR_BASENAME: Final = "xrlint-config"


DEFAULT_CONFIG_FILES: Final = [
    # Added in 0.5.1:
    f"{_REGULAR_BASENAME}.yaml",
    f"{_REGULAR_BASENAME}.yml",
    f"{_REGULAR_BASENAME}.json",
    # Until 0.5.0:
    f"{_MODULE_BASENAME}.yaml",
    f"{_MODULE_BASENAME}.yml",
    f"{_MODULE_BASENAME}.json",
    f"{_MODULE_BASENAME}.py",
]

DEFAULT_CONFIG_FILE_YAML: Final = f"{_REGULAR_BASENAME}.yaml"
DEFAULT_OUTPUT_FORMAT: Final = "simple"
DEFAULT_MAX_WARNINGS: Final = 5

INIT_CONFIG_YAML: Final = (
    "# XRLint configuration file\n"
    "# See https://bcdev.github.io/xrlint/config/\n"
    "\n"
    "- recommended\n"
)

DEFAULT_GLOBAL_FILES: Final = ["**/*.zarr", "**/*.nc"]
DEFAULT_GLOBAL_IGNORES: Final = [".git", "node_modules"]



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

"""
This code example shows how to use the high-level
Python API to validate the contents of an S3 bucket.
"""

import xrlint.all as xrl

URL = "s3://xcube-test/"

xrlint = xrl.XRLint(no_config_lookup=True)
xrlint.init_config("recommended")
results = xrlint.validate_files([URL])
print(xrlint.format_results(results))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import unittest
from dataclasses import dataclass
from typing import Any, Callable, Final, Literal, Type

import xarray as xr

from xrlint.config import ConfigLike
from xrlint.constants import SEVERITY_ERROR
from xrlint.linter import Linter
from xrlint.plugin import new_plugin
from xrlint.result import Message
from xrlint.rule import Rule, RuleMeta, RuleOp
from xrlint.util.formatting import format_count, format_item, format_problems
from xrlint.util.naming import to_snake_case

_PLUGIN_NAME: Final = "testing"


@dataclass(frozen=True, kw_only=True)
class RuleTest:
    """A rule test case."""

    dataset: xr.Dataset
    """The dataset to verify."""

    name: str | None = None
    """A name that helps identifying the test case."""

    args: tuple | list | None = None
    """Optional positional arguments passed to the rule operation's constructor."""

    kwargs: dict[str, Any] | None = None
    """Optional keyword arguments passed to the rule operation's constructor."""

    expected: list[Message | str] | int | None = None
    """Expected messages.
    Either a list of expected message objects or
    the number of expected messages.
    Must not be provided for valid checks
    and must be provided for invalid checks.
    """


class RuleTester:
    """Utility that helps testing rules.

    Args:
        config: Optional configuration-like value.
            For more information see the
            [ConfigLike][xrlint.config.ConfigLike] type alias.
        **config_props: Individual configuration object properties.
            For more information refer to the properties of a
            [ConfigObject][xrlint.config.ConfigObject].
    """

    def __init__(self, *, config: ConfigLike = None, **config_props: Any):
        self._config = config
        self._config_props = config_props

    def run(
        self,
        rule_name: str,
        rule_op_class: Type[RuleOp],
        *,
        valid: list[RuleTest] | None = None,
        invalid: list[RuleTest] | None = None,
    ):
        """Run the given tests in `valid` and `invalid`
        against the given rule.

        Args:
            rule_name: the rule's name
            rule_op_class: a class derived from `RuleOp`
            valid: list of tests that expect no reported problems
            invalid: list of tests that expect reported problems

        Raises:
            AssertionError: if one of the checks fails
        """
        tests = self._create_tests(
            rule_name, rule_op_class, valid=valid, invalid=invalid
        )
        for test in tests.values():
            print(f"Rule {rule_name!r}: running {test.__name__}()...")
            # noinspection PyTypeChecker
            test(None)

    @classmethod
    def define_test(
        cls,
        rule_name: str,
        rule_op_class: Type[RuleOp],
        *,
        valid: list[RuleTest] | None = None,
        invalid: list[RuleTest] | None = None,
        config: ConfigLike = None,
        **config_props: Any,
    ) -> Type[unittest.TestCase]:
        """Create a `unittest.TestCase` class for the given rule and tests.

        The returned class is derived from `unittest.TestCase`
        and contains a test method for each of the tests in
        `valid` and `invalid`.

        Args:
            rule_name: the rule's name
            rule_op_class: the class derived from `RuleOp`
            valid: list of tests that expect no reported problems
            invalid: list of tests that expect reported problems
            config: Optional configuration-like value.
                For more information see the
                [ConfigLike][xrlint.config.ConfigLike] type alias.
            **config_props: Individual configuration object properties.
                For more information refer to the properties of a
                [ConfigObject][xrlint.config.ConfigObject].

        Returns:
            A new class derived from `unittest.TestCase`.
        """
        tester = RuleTester(config=config, **config_props)
        tests = tester._create_tests(
            rule_name, rule_op_class, valid=valid, invalid=invalid
        )
        # noinspection PyTypeChecker
        return type(f"{rule_op_class.__name__}Test", (unittest.TestCase,), tests)

    def _create_tests(
        self,
        rule_name: str,
        rule_op_class: Type[RuleOp],
        valid: list[RuleTest] | None,
        invalid: list[RuleTest] | None,
    ) -> dict[str, Callable[[unittest.TestCase | None], None]]:
        if hasattr(rule_op_class, "meta"):
            tests = dict([self._create_name_test(rule_name, rule_op_class)])
        else:
            tests = {}

        def make_args(checks: list[RuleTest] | None, mode: Literal["valid", "invalid"]):
            return [(check, index, mode) for index, check in enumerate(checks or [])]

        tests |= dict(
            self._create_test(rule_name, rule_op_class, *args)
            for args in make_args(valid, "valid") + make_args(invalid, "invalid")
        )
        return tests

    # noinspection PyMethodMayBeStatic
    def _create_name_test(
        self,
        rule_name: str,
        rule_op_class: Type[RuleOp],
    ) -> tuple[str, Callable]:
        test_id = "test_rule_meta"

        def test_fn(_self: unittest.TestCase):
            rule_meta: RuleMeta = getattr(rule_op_class, "meta")
            assert rule_meta.name == rule_name, (
                f"rule name expected to be {rule_name!r}, but was {rule_meta.name!r}"
            )

        test_fn.__name__ = test_id
        return test_id, test_fn

    def _create_test(
        self,
        rule_name: str,
        rule_op_class: Type[RuleOp],
        test: RuleTest,
        test_index: int,
        test_mode: Literal["valid", "invalid"],
    ) -> tuple[str, Callable]:
        test_id = _format_test_id(test, test_index, test_mode)

        def test_fn(_self: unittest.TestCase):
            error_message = self._test_rule(
                rule_name, rule_op_class, test, test_id, test_mode
            )
            if error_message:
                raise AssertionError(error_message)

        test_fn.__name__ = test_id
        return test_id, test_fn

    def _test_rule(
        self,
        rule_name: str,
        rule_op_class: Type[RuleOp],
        test: RuleTest,
        test_id: str,
        test_mode: Literal["valid", "invalid"],
    ) -> str | None:
        # Note, the rule's code cannot and should not depend
        # on the currently configured severity.
        # There is also no way for a rule to obtain the severity.
        severity = SEVERITY_ERROR
        linter = Linter(self._config, self._config_props)
        result = linter.validate(
            test.dataset,
            plugins={
                _PLUGIN_NAME: (
                    new_plugin(
                        name=_PLUGIN_NAME,
                        rules={
                            rule_name: Rule(
                                meta=RuleMeta(name=rule_name),
                                op_class=rule_op_class,
                            )
                        },
                    )
                )
            },
            rules={
                f"{_PLUGIN_NAME}/{rule_name}": (
                    [severity, *(test.args or ()), (test.kwargs or {})]
                    if test.args or test.kwargs
                    else severity
                )
            },
        )

        result_message_count = len(result.messages)

        expected = test.expected
        expected_message_count = 0
        expected_messages = None

        if test_mode == "valid":
            assert expected is None, (
                f"{test_id}: you cannot provide the keyword argument"
                f" `expected` for a RuleTest in 'valid' mode."
            )
        else:
            if isinstance(expected, int):
                expected_message_count = max(1, expected)
                expected_messages = None
            elif isinstance(expected, list):
                expected_message_count = len(expected)
                expected_messages = expected
            assert expected_message_count > 0, (
                f"{test_id}: you must provide a valid keyword argument"
                f" `expected` for a RuleTest in 'invalid' mode. Pass a list"
                f" of expected message or str objects or an int specifying"
                f" the expected number of messages."
            )

        lines: list[str] = []
        if result_message_count == expected_message_count:
            if expected_messages is None:
                return None
            all_ok = True
            texts = map(_get_message_text, expected_messages)
            result_texts = map(_get_message_text, result.messages)
            for i, (expected_text, result_text) in enumerate(zip(texts, result_texts)):
                if expected_text != result_text:
                    all_ok = False
                    lines.append(f"Message {i}:")
                    lines.append(f"  Expected: {expected_text}")
                    lines.append(f"  Actual: {result_text}")
            if all_ok:
                return None

        else:
            if expected_messages:
                lines.append(
                    f"Expected {format_item(expected_message_count, 'message')}:"
                )
                for i, text in enumerate(map(_get_message_text, expected_messages)):
                    lines.append(f"  {i}: {text}")
            if result.messages:
                lines.append(f"Actual {format_item(result_message_count, 'message')}:")
                for i, text in enumerate(map(_get_message_text, result.messages)):
                    lines.append(f"  {i}: {text}")

        result_text = format_problems(result.error_count, result.warning_count)
        if expected_message_count == result_message_count:
            problem_text = (
                f"got {result_text} as expected, but encountered message mismatch"
            )
        else:
            expected_text = format_count(expected_message_count, "problem")
            problem_text = f"expected {expected_text}, but got {result_text}"

        messages_text = "\n".join(lines)
        messages_text = (":\n" + messages_text) if messages_text else "."
        return f"Rule {rule_name!r}: {test_id}: {problem_text}{messages_text}"


def _format_test_id(
    test: RuleTest, test_index: int, test_mode: Literal["valid", "invalid"]
) -> str:
    if test.name:
        return f"test_{test_mode}_{to_snake_case(test.name)}"
    else:
        return f"test_{test_mode}_{test_index}"


def _get_message_text(m: Message | str) -> str:
    return m if isinstance(m, str) else m.message



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from xrlint.plugins.core.rules.var_units import VarUnits
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset()
valid_dataset_1 = xr.Dataset(
    attrs=dict(title="v-data"),
    coords={"t": xr.DataArray([0, 1, 2], dims="t", attrs={"units": "seconds"})},
    data_vars={"v": xr.DataArray([10, 20, 30], dims="t", attrs={"units": "m/s"})},
)
valid_dataset_2 = valid_dataset_1.copy()
valid_dataset_2.t.encoding["units"] = "seconds since 2025-02-01 12:15:00"
del valid_dataset_2.t.attrs["units"]

valid_dataset_3 = valid_dataset_1.copy()
valid_dataset_3.t.attrs["grid_mapping_name"] = "latitude_longitude"

invalid_dataset_0 = valid_dataset_1.copy()
invalid_dataset_0.t.attrs = {}

invalid_dataset_1 = valid_dataset_1.copy()
invalid_dataset_1.t.attrs = {"units": 1}

invalid_dataset_2 = valid_dataset_1.copy()
invalid_dataset_2.t.attrs = {"units": ""}


VarUnitsTest = RuleTester.define_test(
    "var-units",
    VarUnits,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=["Missing attribute 'units'."]),
        RuleTest(dataset=invalid_dataset_1, expected=["Invalid attribute 'units': 1"]),
        RuleTest(dataset=invalid_dataset_2, expected=["Empty attribute 'units'."]),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any, Literal

from .formatting import format_message_one_of, format_message_type_of

TYPE_NAMES = (
    "null",
    "boolean",
    "integer",
    "number",
    "string",
    "array",
    "object",
)

JsonTypeName = Literal[
    "null",
    "boolean",
    "integer",
    "number",
    "string",
    "array",
    "object",
]

JsonSchema = dict[str, Any] | bool


# noinspection PyPep8Naming
def schema(
    type: JsonTypeName | list[JsonTypeName] | None = None,
    *,
    # common
    default: Any | None = None,
    const: Any | None = None,
    enum: list[Any] | None = None,
    title: str | None = None,
    description: str | None = None,
    # "integer", "number"
    minimum: int | float | None = None,
    maximum: int | float | None = None,
    exclusiveMinimum: int | float | None = None,
    exclusiveMaximum: int | float | None = None,
    # "array"
    items: list[JsonSchema] | JsonSchema | None = None,
    # "object"
    properties: dict[str, JsonSchema] | None = None,
    additionalProperties: bool | None = None,
    required: list[str] | None = None,
) -> JsonSchema:
    """Helper function so you have keyword-arguments for creating schemas."""
    return {
        k: v
        for k, v in dict(
            type=_parse_type(type),
            default=default,
            const=const,
            enum=enum,
            minimum=minimum,
            maximum=maximum,
            exclusiveMinimum=exclusiveMinimum,
            exclusiveMaximum=exclusiveMaximum,
            items=items,
            properties=properties,
            additionalProperties=False if additionalProperties is False else None,
            required=required,
            title=title,
            description=description,
        ).items()
        if v is not None
    }


def _parse_type(type: Any) -> JsonTypeName | list[JsonTypeName] | None:
    if isinstance(type, (list, tuple)):
        if not type:
            return None
        return [_validate_type_name(t) for t in type]
    else:
        if type is None:
            return None
        return _validate_type_name(type)


def _validate_type_name(type_name: Any) -> JsonTypeName:
    if not isinstance(type_name, str):
        raise TypeError(format_message_type_of("type", type_name, "str|list[str]"))
    if type_name not in TYPE_NAMES:
        raise ValueError(format_message_one_of("type name", type_name, TYPE_NAMES))
    # noinspection PyTypeChecker
    return type_name



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.formatter import FormatterRegistry
from xrlint.util.importutil import import_submodules

registry = FormatterRegistry()


def export_formatters() -> FormatterRegistry:
    import_submodules("xrlint.formatters")
    return registry



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import importlib
import pathlib
from typing import Any, Callable, Type, TypeVar

from xrlint.util.formatting import format_message_type_of


def import_submodules(package_name: str, dry_run: bool = False) -> list[str]:
    package = importlib.import_module(package_name)
    assert hasattr(package, "__path__")

    package_path = pathlib.Path(package.__path__[0])

    module_names = []
    for module_file in package_path.iterdir():
        if (
            module_file.is_file()
            and module_file.name.endswith(".py")
            and module_file.name != "__init__.py"
        ):
            module_names.append(module_file.name[:-3])
        elif (
            module_file.is_dir()
            and module_file.name != "__pycache__"
            and (module_file / "__init__.py").is_file()
        ):
            module_names.append(module_file.name)

    qual_module_names = [f"{package_name}.{m}" for m in module_names]

    if not dry_run:
        for qual_module_name in qual_module_names:
            importlib.import_module(qual_module_name)

    return qual_module_names


T = TypeVar("T")


def import_value(
    module_ref: str,
    attr_ref: str | None = None,
    *,
    constant: bool = False,
    factory: Callable[[Any], T] | None = None,
    expected_type: Type[T] | None = None,
) -> tuple[T, str]:
    """Import an exported value from given module reference.

    Args:
        module_ref: Module reference. A string comprising either a fully
            qualified module name plus an optional attribute reference
            using format "<module-name>:<attr-ref>" or just a module name.
            In this case, `attr_ref` should be given.
            If it is not given, the module itself will be the exported value.
        attr_ref: Attribute reference. Should be given in the case where
            `module_ref` does not contain an attribute reference.
            Example values are "export_plugin", "export_config".
        constant: If `True` the value is expected to be a constant.
            If `False`, the default, the referenced attribute can
            be a no-arg callable that yields the actual exported value.
        factory: 1-arg factory function that converts a value of unknown
            type into `T`. Optional.
        expected_type: The expected value type that is a `T`. Optional.

    Returns:
        value: The imported value of type `T`.
        value_ref: The reference from which the value was imported.

    Raises:
        ValueImportError: if the value could not be imported
    """
    if ":" in module_ref:
        module_name, attr_ref = module_ref.rsplit(":", maxsplit=1)
    else:
        module_name = module_ref
        if attr_ref:
            module_ref = f"{module_name}:{attr_ref}"

    try:
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise ValueImportError(
            f"failed to import value from {module_ref!r}: {e}"
        ) from e

    attr_value = module
    if attr_ref:
        attr_names = attr_ref.split(".")
        for i, attr_name in enumerate(attr_names):
            try:
                attr_value = getattr(attr_value, attr_name)
            except AttributeError:
                raise ValueImportError(
                    f"attribute {'.'.join(attr_names[: i + 1])!r}"
                    f" not found in module {module_name!r}"
                )

    should_invoke = not constant and callable(attr_value)
    if should_invoke:
        # We don't catch exceptions here,
        # because they occur in user land.
        # noinspection PyCallingNonCallable
        exported_value = attr_value()
    else:
        exported_value = attr_value

    if factory is not None:
        try:
            exported_value = factory(exported_value)
        except (ValueError, TypeError) as e:
            raise ValueImportError(
                f"failed converting value of {module_ref!r}: {e}"
            ) from e

    if expected_type is not None and not isinstance(exported_value, expected_type):
        raise ValueImportError(
            format_message_type_of(
                f"value of {module_ref}{('()' if should_invoke else '')}",
                exported_value,
                expected_type,
            )
        )

    return exported_value, module_ref


class ValueImportError(ImportError):
    """Special error that is raised while
    importing an exported value.
    """



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).
import numpy as np
import xarray as xr

from xrlint.plugins.core.rules.var_missing_data import VarMissingData
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset()
valid_dataset_1 = xr.Dataset(
    attrs=dict(title="v-data"),
    coords={"t": xr.DataArray([0, 1, 2], dims="t", attrs={"units": "seconds"})},
    data_vars={"v": xr.DataArray([10, 20, 30], dims="t", attrs={"units": "m/s"})},
)

invalid_dataset_0 = valid_dataset_1.copy(deep=True)
invalid_dataset_0.t.attrs["_FillValue"] = -999

invalid_dataset_1 = valid_dataset_1.copy(deep=True)
invalid_dataset_1.t.encoding["_FillValue"] = -999

invalid_dataset_2 = valid_dataset_1.copy(deep=True)
invalid_dataset_2.v.attrs["scaling_factor"] = 0.01

invalid_dataset_3 = valid_dataset_1.copy(deep=True)
invalid_dataset_3.v.encoding["dtype"] = np.dtype(np.float64)

invalid_dataset_4 = valid_dataset_1.copy(deep=True)
invalid_dataset_4.v.attrs["valid_range"] = [0, 1]

VarMissingDataTest = RuleTester.define_test(
    "var-missing-data",
    VarMissingData,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
    ],
    invalid=[
        RuleTest(
            dataset=invalid_dataset_0,
            expected=[
                "Unexpected attribute '_FillValue', coordinates must not have missing data."
            ],
        ),
        RuleTest(
            dataset=invalid_dataset_1,
            expected=[
                "Unexpected encoding '_FillValue', coordinates must not have missing data."
            ],
        ),
        RuleTest(
            dataset=invalid_dataset_2,
            expected=["Missing attribute '_FillValue' since data packing is used."],
        ),
        RuleTest(
            dataset=invalid_dataset_3,
            expected=["Missing attribute '_FillValue', which should be NaN."],
        ),
        RuleTest(
            dataset=invalid_dataset_4,
            expected=["Valid ranges are not recognized by xarray (as of Feb 2025)."],
        ),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.core.rules.var_flags import VarFlags
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset()
valid_dataset_1 = xr.Dataset(
    attrs=dict(title="sensor-data"),
    data_vars={
        "sensor_status_qc": xr.DataArray(
            [1, 3, 5, 2, 0, 5],
            dims="x",
            attrs=dict(
                long_name="Sensor Status",
                standard_name="status_flag",
                _FillValue=0,
                valid_range=[1, 15],
                flag_masks=[1, 2, 12, 12, 12],
                flag_values=[1, 2, 4, 8, 12],
                flag_meanings=(
                    "low_battery"
                    " hardware_fault"
                    " offline_mode"
                    " calibration_mode"
                    " maintenance_mode"
                ),
            ),
        )
    },
)

# Valid, because and flag_values and flag_meanings are sufficient
valid_dataset_2 = valid_dataset_1.copy()
del valid_dataset_2.sensor_status_qc.attrs["flag_masks"]

# Valid, because and flag_masks and flag_meanings are sufficient
valid_dataset_3 = valid_dataset_1.copy()
del valid_dataset_3.sensor_status_qc.attrs["flag_values"]

# Invalid, because flag_meanings are needed
invalid_dataset_0 = valid_dataset_1.copy()
del invalid_dataset_0.sensor_status_qc.attrs["flag_meanings"]

# Invalid, because flag_values are invalid
invalid_dataset_1 = valid_dataset_1.copy()
invalid_dataset_1.sensor_status_qc.attrs["flag_values"] = "1, 2, 4, 8, 12"

# Invalid, because flag_masks are invalid
invalid_dataset_2 = valid_dataset_1.copy()
invalid_dataset_2.sensor_status_qc.attrs["flag_masks"] = "1, 2, 12, 12, 12"

# Invalid, because flag_values and flag_masks must have same length
invalid_dataset_3 = valid_dataset_1.copy()
invalid_dataset_3.sensor_status_qc.attrs["flag_masks"] = [1, 2, 12]

# Invalid, because missing flag_values and flag_masks
invalid_dataset_4 = valid_dataset_1.copy()
del invalid_dataset_4.sensor_status_qc.attrs["flag_values"]
del invalid_dataset_4.sensor_status_qc.attrs["flag_masks"]

# Invalid, because flag_meanings type is not ok
invalid_dataset_5 = valid_dataset_1.copy()
invalid_dataset_5.sensor_status_qc.attrs["flag_meanings"] = [1, 2, 12, 12, 12]

# Invalid, because flag_meanings length is not ok
invalid_dataset_6 = valid_dataset_1.copy()
invalid_dataset_6.sensor_status_qc.attrs["flag_meanings"] = "a b c"

# Invalid, because flag variable is not int
invalid_dataset_7 = valid_dataset_1.copy()
invalid_dataset_7["sensor_status_qc"] = valid_dataset_1.sensor_status_qc.astype(
    np.float64
)

VarFlagsTest = RuleTester.define_test(
    "var-flags",
    VarFlags,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=2),
        RuleTest(dataset=invalid_dataset_1, expected=1),
        RuleTest(dataset=invalid_dataset_2, expected=1),
        RuleTest(dataset=invalid_dataset_3, expected=1),
        RuleTest(dataset=invalid_dataset_4, expected=1),
        RuleTest(dataset=invalid_dataset_5, expected=1),
        RuleTest(dataset=invalid_dataset_6, expected=1),
        RuleTest(dataset=invalid_dataset_7, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.constants import ML_META_FILENAME
from xrlint.plugins.xcube.plugin import plugin
from xrlint.plugins.xcube.util import get_dataset_level_info, is_spatial_var
from xrlint.rule import RuleContext, RuleOp
from xrlint.util.formatting import format_item


@plugin.define_rule(
    "ml-dataset-meta",
    version="1.0.0",
    type="suggestion",
    description=(
        f"Multi-level datasets should provide a {ML_META_FILENAME!r}"
        f" meta-info file, and if so, it should be consistent."
        f" Without the meta-info file the multi-level dataset cannot be"
        f" reliably extended by new time slices as the aggregation method"
        f" used for each variable must be specified."
    ),
    docs_url=(
        "https://xcube.readthedocs.io/en/latest/mldatasets.html#the-xcube-levels-format"
    ),
)
class MLDatasetMeta(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        level_info = get_dataset_level_info(node.dataset)
        if level_info is None:
            # ok, this rules applies only to level datasets opened
            # by the xcube multi-level processor
            return

        level = level_info.level
        if level > 0:
            # ok, this rule does only apply to level 0
            return

        meta = level_info.meta
        if meta is None:
            ctx.report(
                f"Missing {ML_META_FILENAME!r} meta-info file.",
                suggestions=[
                    f"Add {ML_META_FILENAME!r} meta-info file."
                    f" Without the meta-info the dataset cannot be reliably extended"
                    f" as the aggregation method used for each variable must be"
                    f" specified."
                ],
            )
            return

        if not meta.version.startswith("1."):
            ctx.report(f"Unsupported {ML_META_FILENAME!r} meta-info version.")

        if meta.num_levels <= 0:
            ctx.report(
                f"Invalid 'num_levels' in {ML_META_FILENAME!r} meta-info:"
                f" {meta.num_levels}."
            )
        elif meta.num_levels != level_info.num_levels:
            ctx.report(
                f"Expected {format_item(meta.num_levels, 'level')},"
                f" but found {level_info.num_levels}."
            )

        if meta.use_saved_levels is None:
            ctx.report(
                f"Missing value for 'use_saved_levels'"
                f" in {ML_META_FILENAME!r} meta-info."
            )

        if not meta.agg_methods:
            ctx.report(
                f"Missing value for 'agg_methods' in {ML_META_FILENAME!r} meta-info."
            )
        else:
            for var_name, var in node.dataset.data_vars.items():
                if is_spatial_var(var) and not meta.agg_methods.get(var_name):
                    ctx.report(
                        f"Missing value for variable {var_name!r}"
                        f" in 'agg_methods' of {ML_META_FILENAME!r} meta-info."
                    )
            for var_name in meta.agg_methods.keys():
                if var_name not in node.dataset:
                    ctx.report(
                        f"Variable {var_name!r} not found in dataset, but specified"
                        f" in 'agg_methods' of {ML_META_FILENAME!r} meta-info."
                    )

        # Later: check meta.tile_size as well...



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from tests.plugins.xcube.helpers import make_cube_levels
from xrlint.plugins.xcube.rules.ml_dataset_meta import MLDatasetMeta
from xrlint.plugins.xcube.util import (
    LevelInfo,
    LevelsMeta,
    get_dataset_level_info,
    set_dataset_level_info,
)
from xrlint.testing import RuleTest, RuleTester


def _replace_meta(dataset: xr.Dataset, meta: LevelsMeta) -> xr.Dataset:
    dataset = dataset.copy()
    old_level_info = get_dataset_level_info(dataset)
    new_level_info = LevelInfo(
        level=old_level_info.level,
        num_levels=old_level_info.num_levels,
        datasets=old_level_info.datasets,
        meta=meta,
    )
    set_dataset_level_info(dataset, new_level_info)
    return dataset


levels_with_meta = make_cube_levels(
    4,
    720,
    360,
    meta=LevelsMeta(
        version="1.0",
        num_levels=4,
        use_saved_levels=True,
        agg_methods={"chl": "mean"},
    ),
)

valid_dataset_0 = levels_with_meta[0]
valid_dataset_1 = levels_with_meta[1]
valid_dataset_2 = levels_with_meta[2]
valid_dataset_3 = xr.Dataset()

levels_wo_meta = make_cube_levels(4, 720, 360, force_infos=True)
invalid_dataset_0 = levels_wo_meta[0]
invalid_dataset_1 = _replace_meta(
    levels_wo_meta[0].copy(),
    meta=LevelsMeta(
        version="2.0",  # error: != "1.x"
        num_levels=0,  # error: < 1
        # error: missing use_saved_levels=False
        # error: missing agg_methods={"chl": "mean"}
    ),
)
invalid_dataset_2 = _replace_meta(
    levels_wo_meta[0],
    meta=LevelsMeta(
        version="1.0",  # ok
        num_levels=3,  # error: != level_info.num_levels
    ),
)

invalid_dataset_3 = _replace_meta(
    levels_wo_meta[0],
    meta=LevelsMeta(
        version="1.0",  # ok
        num_levels=4,  # ok
        use_saved_levels=False,  # ok
        agg_methods={"tsm": "median"},  # error: where is "chl"?
    ),
)

MLDatasetMetaTest = RuleTester.define_test(
    "ml-dataset-meta",
    MLDatasetMeta,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=1),
        RuleTest(dataset=invalid_dataset_1, expected=4),
        RuleTest(dataset=invalid_dataset_2, expected=3),
        RuleTest(dataset=invalid_dataset_3, expected=2),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import json
import os
from collections.abc import Iterable, Iterator

import click
import fsspec
import fsspec.core
import fsspec.implementations.local
import yaml

from xrlint.cli.config import ConfigError, read_config
from xrlint.cli.constants import (
    DEFAULT_CONFIG_FILE_YAML,
    DEFAULT_CONFIG_FILES,
    DEFAULT_GLOBAL_FILES,
    DEFAULT_GLOBAL_IGNORES,
    DEFAULT_MAX_WARNINGS,
    DEFAULT_OUTPUT_FORMAT,
    INIT_CONFIG_YAML,
)
from xrlint.config import Config, ConfigLike, ConfigObject, get_core_config_object
from xrlint.formatter import FormatterContext
from xrlint.formatters import export_formatters
from xrlint.linter import Linter
from xrlint.plugin import Plugin
from xrlint.result import Result, ResultStats
from xrlint.util.filefilter import FileFilter

DEFAULT_GLOBAL_FILTER = FileFilter.from_patterns(
    DEFAULT_GLOBAL_FILES, DEFAULT_GLOBAL_IGNORES
)


class XRLint(FormatterContext):
    """This class provides the engine behind the XRLint
    CLI application.
    It represents the highest level component in the Python API.
    """

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        no_config_lookup: int = False,
        config_path: str | None = None,
        plugin_specs: tuple[str, ...] = (),
        rule_specs: tuple[str, ...] = (),
        output_format: str = DEFAULT_OUTPUT_FORMAT,
        output_path: str | None = None,
        output_styled: bool = True,
        max_warnings: int = DEFAULT_MAX_WARNINGS,
    ):
        self.no_config_lookup = no_config_lookup
        self.config_path = config_path
        self.plugin_specs = plugin_specs
        self.rule_specs = rule_specs
        self.output_format = output_format
        self.output_path = output_path
        self.output_styled = output_styled
        self.max_warnings = max_warnings
        self._result_stats = ResultStats()
        self.config = Config()

    @property
    def max_warnings_exceeded(self) -> bool:
        """`True` if the maximum number of warnings has been exceeded."""
        return self._result_stats.warning_count > self.max_warnings

    @property
    def result_stats(self) -> ResultStats:
        """Get current result statistics."""
        return self._result_stats

    def init_config(self, *extra_configs: ConfigLike) -> None:
        """Initialize configuration.
        The function will load the configuration list from a specified
        configuration file, if any.
        Otherwise, it will search for the default configuration files
        in the current working directory.

        Args:
            *extra_configs: Variable number of configuration-like arguments.
                For more information see the
                [ConfigLike][xrlint.config.ConfigLike] type alias.
                If provided, `extra_configs` will be appended to configuration
                read from configration files and passed as command line options.
        """
        plugins = {}
        for plugin_spec in self.plugin_specs:
            plugin = Plugin.from_value(plugin_spec)
            plugins[plugin.meta.name] = plugin

        rules = {}
        for rule_spec in self.rule_specs:
            rule = yaml.load(rule_spec, Loader=yaml.SafeLoader)
            rules.update(rule)

        file_config = None

        if self.config_path:
            try:
                file_config = read_config(self.config_path)
            except (FileNotFoundError, ConfigError) as e:
                raise click.ClickException(f"{e}") from e
        elif not self.no_config_lookup:
            for config_path in DEFAULT_CONFIG_FILES:
                try:
                    file_config = read_config(config_path)
                    break
                except FileNotFoundError:
                    pass
                except ConfigError as e:
                    raise click.ClickException(f"{e}") from e
            if file_config is None:
                click.echo("Warning: no configuration file found.")

        core_config_obj = get_core_config_object()
        core_config_obj.plugins.update(plugins)

        base_configs = []
        if file_config is not None:
            base_configs += file_config.objects
        if rules:
            base_configs += [{"rules": rules}]

        self.config = Config.from_config(core_config_obj, *base_configs, *extra_configs)
        if not any(co.rules for co in self.config.objects):
            raise click.ClickException("no rules configured")

    def compute_config_for_file(self, file_path: str) -> ConfigObject | None:
        """Compute the configuration object for the given file.

        Args:
            file_path: A file path or URL.

        Returns:
            A configuration object or `None` if no item
                in the configuration list applies.
        """
        return self.config.compute_config_object(file_path)

    def print_config_for_file(self, file_path: str) -> None:
        """Print computed configuration object for the given file.

        Args:
            file_path: A file path or URL.
        """
        config = self.compute_config_for_file(file_path)
        config_json_obj = config.to_json() if config is not None else None
        click.echo(json.dumps(config_json_obj, indent=2))

    def validate_files(self, files: Iterable[str]) -> Iterator[Result]:
        """Validate given files or directories which may also be given as URLs.
        The function produces a validation result for each file.

        Args:
            files: Iterable of files.

        Returns:
            Iterator of reports.
        """
        linter = Linter()
        for file_path, config in self.get_files(files):
            yield linter.validate(file_path, config=config)

    def get_files(
        self, file_paths: Iterable[str]
    ) -> Iterator[tuple[str, ConfigObject]]:
        """Provide an iterator for the list of files or directories and their
        computed configurations.

        Directories in `files` that are not ignored and not recognized by any
        file pattern will be recursively traversed.

        Args:
            file_paths: Iterable of files or directories.

        Returns:
            An iterator of pairs comprising a file or directory path
                and its computed configuration.
        """
        config, global_filter = self.config.split_global_filter(
            default=DEFAULT_GLOBAL_FILTER
        )

        def compute_config_object(p: str):
            return config.compute_config_object(p) if global_filter.accept(p) else None

        for file_path in file_paths:
            _fs, root = fsspec.url_to_fs(file_path)
            fs: fsspec.AbstractFileSystem = _fs
            is_local = isinstance(fs, fsspec.implementations.local.LocalFileSystem)

            config_obj = compute_config_object(file_path)
            if config_obj is not None:
                yield file_path, config_obj
                continue

            if fs.isdir(root):
                for path, dirs, files in fs.walk(root, topdown=True):
                    for d in list(dirs):
                        d_path = f"{path}/{d}"
                        d_path = d_path if is_local else fs.unstrip_protocol(d_path)
                        c = compute_config_object(d_path)
                        if c is not None:
                            dirs.remove(d)
                            yield d_path, c

                    for f in files:
                        f_path = f"{path}/{f}"
                        f_path = f_path if is_local else fs.unstrip_protocol(f_path)
                        c = compute_config_object(f_path)
                        if c is not None:
                            yield f_path, c

    def format_results(self, results: Iterable[Result]) -> str:
        """Format the given results.

        Args:
            results: Iterable of results.

        Returns:
            A report in plain text.
        """
        output_format = (
            self.output_format if self.output_format else DEFAULT_OUTPUT_FORMAT
        )
        formatters = export_formatters()
        formatter = formatters.get(output_format)
        if formatter is None:
            raise click.ClickException(
                f"unknown format {output_format!r}."
                f" The available formats are"
                f" {', '.join(repr(k) for k in formatters.keys())}."
            )
        # Here we could pass and validate format-specific args/kwargs
        #   against formatter.meta.schema
        if output_format == "simple":
            formatter_kwargs = {
                "styled": self.output_styled and self.output_path is None,
                "output": self.output_path is None,
            }
        else:
            formatter_kwargs = {}
        # noinspection PyArgumentList
        formatter_op = formatter.op_class(**formatter_kwargs)
        return formatter_op.format(self, self._result_stats.collect(results))

    def write_report(self, report: str) -> None:
        """Write the validation report provided as plain text."""
        if self.output_path:
            with fsspec.open(self.output_path, mode="w") as f:
                f.write(report)
        elif self.output_format != "simple":
            # The simple formatters outputs incrementally to console
            print(report)

    @classmethod
    def init_config_file(cls) -> None:
        """Write an initial configuration file.
        The file is written into the current working directory.
        """
        file_path = DEFAULT_CONFIG_FILE_YAML
        if os.path.exists(file_path):
            raise click.ClickException(f"file {file_path} already exists.")
        with open(file_path, "w") as f:
            f.write(INIT_CONFIG_YAML)
        click.echo(f"Configuration template written to {file_path}")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import unittest

from xrlint.util.filefilter import FileFilter
from xrlint.util.filepattern import FilePattern


class FileFilterTest(unittest.TestCase):
    def test_constructor(self):
        file_filter = FileFilter()
        self.assertEqual((), file_filter.files)
        self.assertEqual((), file_filter.ignores)

        files = (FilePattern("**/*.zarr"),)
        ignores = (FilePattern(".git"),)
        file_filter = FileFilter(files, ignores)
        self.assertIs(files, file_filter.files)
        self.assertIs(ignores, file_filter.ignores)

    def test_empty(self):
        file_filter = FileFilter()
        self.assertEqual(True, file_filter.empty)

        file_filter = FileFilter((), ())
        self.assertEqual(True, file_filter.empty)

        file_filter = FileFilter((), (FilePattern(".git"),))
        self.assertEqual(False, file_filter.empty)

        file_filter = FileFilter((FilePattern("**/*.zarr"),), ())
        self.assertEqual(False, file_filter.empty)

        file_filter = FileFilter((FilePattern("**/*.zarr"),), (FilePattern(".git"),))
        self.assertEqual(False, file_filter.empty)

    def test_from_patterns(self):
        self.assertEqual(FileFilter(), FileFilter.from_patterns(None, None))
        self.assertEqual(FileFilter(), FileFilter.from_patterns([], []))
        self.assertEqual(
            FileFilter((FilePattern("**/*.zarr"),), (FilePattern(".git"),)),
            FileFilter.from_patterns(["**/*.zarr"], [".git"]),
        )

    def test_merge(self):
        file_filter_1 = FileFilter((FilePattern("**/*.zarr"),), (FilePattern(".git"),))
        file_filter_2 = FileFilter((FilePattern("**/*.nc"),), (FilePattern("src/"),))
        self.assertEqual(
            FileFilter(
                (FilePattern("**/*.zarr"), FilePattern("**/*.nc")),
                (FilePattern(".git"), FilePattern("src/")),
            ),
            file_filter_1.merge(file_filter_2),
        )

    def test_accept(self):
        file_filter = FileFilter.from_patterns(["**/*.zarr"], None)
        self.assertEqual(True, file_filter.accept("test.zarr"))
        self.assertEqual(False, file_filter.accept("test.nc"))
        self.assertEqual(True, file_filter.accept("./temp/x.zarr"))

        file_filter = FileFilter.from_patterns(["**/*.zarr"], ["**/test.zarr"])
        self.assertEqual(False, file_filter.accept("test.zarr"))
        self.assertEqual(True, file_filter.accept("test-1.zarr"))
        self.assertEqual(True, file_filter.accept("test-2.zarr"))

        # negated ignores:
        file_filter = FileFilter.from_patterns(
            ["**/*.zarr"], ["**/temp/*", "!**/temp/x.zarr"]
        )
        self.assertEqual(True, file_filter.accept("./test.zarr"))
        self.assertEqual(True, file_filter.accept("./temp/x.zarr"))
        self.assertEqual(False, file_filter.accept("./temp/y.zarr"))

        # negated ignores:
        file_filter = FileFilter.from_patterns([], ["**/temp/*", "!**/temp/*.zarr"])
        self.assertEqual(True, file_filter.accept("./ds.nc"))
        self.assertEqual(True, file_filter.accept("./temp/ds.zarr"))
        self.assertEqual(True, file_filter.accept("./temp/test/ds.zarr"))
        self.assertEqual(False, file_filter.accept("./temp/ds.nc"))
        self.assertEqual(False, file_filter.accept("./temp/README.md"))

        # just ignores:
        file_filter = FileFilter.from_patterns([], ["**/*.json", "**/*.yaml"])
        self.assertEqual(True, file_filter.accept("./test.nc"))
        self.assertEqual(True, file_filter.accept("./temp/x.zarr"))
        self.assertEqual(False, file_filter.accept("c.yaml"))
        self.assertEqual(False, file_filter.accept("test/config.json"))
        self.assertEqual(False, file_filter.accept("test/config.yaml"))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.xcube.rules.data_var_colors import DataVarColors
from xrlint.testing import RuleTest, RuleTester


def make_dataset():
    dims = ["time", "y", "x"]
    n = 5
    return xr.Dataset(
        attrs=dict(title="v-data"),
        coords={
            dims[2]: xr.DataArray(
                np.linspace(0, 1, n), dims=dims[2], attrs={"units": "m"}
            ),
            dims[1]: xr.DataArray(
                np.linspace(0, 1, n), dims=dims[1], attrs={"units": "m"}
            ),
            dims[0]: xr.DataArray(
                [2010, 2011, 2012, 2013, 2014],
                dims=dims[0],
                attrs={"units": "years"},
            ),
        },
        data_vars={
            "chl": xr.DataArray(
                np.random.random((n, n, n)),
                dims=dims,
                attrs={
                    "units": "mg/m^-3",
                    "color_bar_name": "plasma",
                    "color_value_min": 0,
                    "color_value_max": 100,
                    "color_norm": "log",
                },
            ),
        },
    )


valid_dataset_1 = make_dataset()

invalid_dataset_1 = make_dataset()
invalid_dataset_1.chl.attrs = {
    "units": "mg/m^-3",
    # Missing:
    # "color_bar_name": "plasma",
    # "color_value_min": 0,
    # "color_value_max": 100,
    # "color_norm": "log",
}
invalid_dataset_2 = make_dataset()
invalid_dataset_2.chl.attrs = {
    "units": "mg/m^-3",
    "color_bar_name": "plasma",
    # Missing:
    # "color_value_min": 0,
    # "color_value_max": 100,
    # "color_norm": "log",
}
invalid_dataset_3 = make_dataset()
invalid_dataset_3.chl.attrs = {
    "units": "mg/m^-3",
    "color_bar_name": "plasma",
    "color_value_min": 0,
    "color_value_max": 100,
    "color_norm": "ln",  # wrong
}

LatLonNamingTest = RuleTester.define_test(
    "data-var-colors",
    DataVarColors,
    valid=[
        RuleTest(dataset=valid_dataset_1),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_1, expected=1),
        RuleTest(dataset=invalid_dataset_2, expected=1),
        RuleTest(dataset=invalid_dataset_3, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from collections.abc import Sequence
from typing import Any


def format_problems(error_count: int, warning_count: int) -> str:
    """Return human readable text for the given
    `error_count` and `warning_count`.
    """
    problem_count = error_count + warning_count
    p_label = format_count(problem_count, "problem")
    if problem_count == 0:
        return p_label
    e_label = format_count(error_count, "error")
    w_label = format_count(warning_count, "warning")
    if error_count and warning_count:
        return f"{p_label} ({e_label} and {w_label})"
    if error_count:
        return e_label
    else:
        return w_label


def format_count(
    count: int | float,
    singular: str,
    plural: str | None = None,
    upper: bool | None = None,
) -> str:
    """Format given `count` of items named by `singular` or `plural`."""
    if count == 0:
        count_text = format_case("no", upper)
    elif count == 1:
        count_text = format_case("one", upper)
    else:
        count_text = str(count)
    return f"{count_text} {format_item(count, singular, plural=plural)}"


def format_item(
    count: int | float,
    singular: str,
    plural: str | None = None,
    upper: bool | None = None,
) -> str:
    """Format `singular` given it occurs `count` times."""
    if count == 1:
        name_text = singular
    else:
        name_text = plural or (singular + "s")
    return format_case(name_text, upper)


def format_case(text: str, upper: bool | None = None) -> str:
    """Return `text` with first character turned to uppercase or lowercase."""
    if not text or upper is None:
        return text
    return (text[0].upper() if upper else text[0].lower()) + text[1:]


def format_message_one_of(name: str, value: Any, enum_value) -> str:
    if isinstance(enum_value, str):
        enum_text = enum_value
    else:
        enum_text = ", ".join(f"{v!r}" for v in enum_value)
    return f"{name} must be one of {enum_text}, but got {value!r}"


def format_message_type_of(name: str, value: Any, type_value: type | str) -> str:
    return (
        f"{name} must be of type {format_type_of(type_value)},"
        f" but got {format_type_of(type(value))}"
    )


def format_type_of(value: Any) -> str:
    if value is None or value is type(None):
        return "None"
    if isinstance(value, str):
        return value
    assert isinstance(value, type)
    return value.__name__


def format_seq(seq: Sequence, max_count: int = 6) -> str:
    if len(seq) == 0:
        return ""
    elif len(seq) <= max_count:
        return _format_seq(seq)
    else:
        i = max_count // 2
        return f"{_format_seq(seq[:i])}, ..., {_format_seq(seq[-i:])}"


def _format_seq(seq: Sequence) -> str:
    return ", ".join(str(v) for v in seq)


def format_styled(
    text: str = "", s: str = "", fg: str = "", bg: str = "", href: str = ""
):
    """Format styled text"""
    if not text:
        if not href:
            return ""
        text = href

    style = ""
    if s != "":
        style += _S_CODES.get(s, "")
    if fg != "":
        style += ";" + _FG_CODES.get(fg, "")
    if bg != "":
        style += (";" if style else ";;") + _BG_CODES.get(bg, "")

    if style:
        styled_text = f"\033[{style}m{text}\033[0m"
    else:
        styled_text = text

    if not href:
        return styled_text
    if "://" in href:
        url = href
    else:
        url = "file://" + href
    return f"\033]8;;{url}\033\\{styled_text}\033]8;;\033\\"


_S_CODES = {
    k: str(c)
    for k, c in (
        ("normal", 0),
        ("bold", 1),
        ("dim", 2),
        ("italic", 3),
        ("underline", 4),
        ("blink", 5),
        ("reverse", 7),
    )
}

_C_CODES = (
    ("black", 30, 40),
    ("red", 31, 41),
    ("green", 32, 42),
    ("yellow", 33, 43),
    ("blue", 34, 44),
    ("magenta", 35, 45),
    ("cyan", 36, 46),
    ("white", 37, 47),
)

_FG_CODES = {k: str(c) for k, c, _ in _C_CODES}
_BG_CODES = {k: str(c) for k, _, c in _C_CODES}



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

import pytest
import xarray as xr

# noinspection PyProtectedMember
from xrlint._linter.rulectx import RuleContextImpl
from xrlint.config import ConfigObject
from xrlint.constants import DATASET_ROOT_NAME
from xrlint.node import DatasetNode
from xrlint.plugins.core.rules.access_latency import AccessLatency
from xrlint.result import Message
from xrlint.rule import RuleExit

valid_dataset_0 = xr.Dataset()

invalid_dataset_0 = xr.Dataset()


class OpeningTimeTest(TestCase):
    @classmethod
    def invoke_op(
        cls, dataset: xr.Dataset, access_latency: float, threshold: float | None = None
    ):
        ctx = RuleContextImpl(
            config=ConfigObject(),
            dataset=dataset,
            file_path="test.zarr",
            file_index=None,
            access_latency=access_latency,
        )
        node = DatasetNode(
            parent=None,
            path=DATASET_ROOT_NAME,
            name=DATASET_ROOT_NAME,
            dataset=ctx.dataset,
        )
        rule_op = (
            AccessLatency(threshold=threshold)
            if threshold is not None
            else AccessLatency()
        )
        with pytest.raises(RuleExit):
            rule_op.validate_dataset(ctx, node)
        return ctx

    def test_valid(self):
        ctx = self.invoke_op(xr.Dataset(), 1.0, threshold=None)
        self.assertEqual([], ctx.messages)

        ctx = self.invoke_op(xr.Dataset(), 1.0, threshold=1.0)
        self.assertEqual([], ctx.messages)

    def test_invalid(self):
        ctx = self.invoke_op(xr.Dataset(), 3.16, threshold=None)
        self.assertEqual(
            [
                Message(
                    message="Access latency exceeds threshold: 3.2 > 2.5 seconds.",
                    node_path=DATASET_ROOT_NAME,
                    severity=2,
                )
            ],
            ctx.messages,
        )

        ctx = self.invoke_op(xr.Dataset(), 0.2032, threshold=0.1)
        self.assertEqual(
            [
                Message(
                    message="Access latency exceeds threshold: 0.2 > 0.1 seconds.",
                    node_path=DATASET_ROOT_NAME,
                    severity=2,
                )
            ],
            ctx.messages,
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import json
from collections.abc import Iterable

from xrlint.formatter import FormatterContext, FormatterOp
from xrlint.formatters import registry
from xrlint.result import Result, get_rules_meta_for_results
from xrlint.util.schema import schema


@registry.define_formatter(
    "json",
    version="1.0.0",
    schema=schema(
        "object",
        properties=dict(
            indent=schema("integer", minimum=0, maximum=8, default=2),
            with_meta=schema("boolean", default=False),
        ),
    ),
)
class Json(FormatterOp):
    def __init__(self, indent: int = 2, with_meta: bool = False):
        super().__init__()
        self.indent = indent
        self.with_meta = with_meta

    def format(
        self,
        context: FormatterContext,
        results: Iterable[Result],
    ) -> str:
        results = list(results)  # get them all

        omitted_props = {"config"}
        results_json = {
            "results": [
                {k: v for k, v in r.to_json().items() if k not in omitted_props}
                for r in results
            ],
        }
        if self.with_meta:
            rules_meta = get_rules_meta_for_results(results)
            results_json.update(
                {
                    "rules_meta": [rm.to_json() for rm in rules_meta.values()],
                }
            )
        return json.dumps(results_json, indent=self.indent)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from dataclasses import dataclass, field
from types import NoneType, UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Mapping,
    Optional,
    TypeAlias,
    Union,
    get_args,
    get_origin,
)
from unittest import TestCase

import pytest

from xrlint.util.constructible import (
    MappingConstructible,
    ValueConstructible,
    get_class_parameters,
)
from xrlint.util.serializable import JsonSerializable


@dataclass()
class UselessContainer(ValueConstructible):
    pass


@dataclass()
class RequiredPropsContainer(MappingConstructible):
    x: float
    y: float
    z: float


class NoTypesContainer(MappingConstructible):
    def __init__(self, u, v, w, /):  # positional only!
        self.u = u
        self.v = v
        self.w = w


@dataclass()
class SimpleTypesContainer(MappingConstructible, JsonSerializable):
    a: Any = None
    b: bool = False
    c: int = 0
    d: float = 0.0
    e: str = "abc"
    f: type = int
    g: Literal[0, 1, True, False, "on", "off"] = 0


@dataclass()
class ComplexTypesContainer(MappingConstructible, JsonSerializable):
    p: SimpleTypesContainer = field(default_factory=SimpleTypesContainer)
    q: dict[str, bool] = field(default_factory=dict)
    r: dict[str, SimpleTypesContainer] = field(default_factory=dict)
    s: list[int] = field(default_factory=list)
    t: list[SimpleTypesContainer] = field(default_factory=list)
    u: int | float | None = None


@dataclass()
class UnionTypesContainer(MappingConstructible, JsonSerializable):
    m: SimpleTypesContainer | ComplexTypesContainer | None = None


if TYPE_CHECKING:
    # make IDEs and flake8 happy
    from xrlint.plugin import Plugin
    from xrlint.rule import RuleConfig


@dataclass()
class UnresolvedTypesContainer(ComplexTypesContainer, SimpleTypesContainer):
    rules: dict[str, "RuleConfig"] = field(default_factory=dict)
    plugins: dict[str, "Plugin"] = field(default_factory=dict)

    @classmethod
    def forward_refs(cls) -> Optional[Mapping[str, type]]:
        from xrlint.plugin import Plugin
        from xrlint.rule import RuleConfig

        return {
            "RuleConfig": RuleConfig,
            "Plugin": Plugin,
        }


T1: TypeAlias = int | str | Union[bool, None] | None
T2: TypeAlias = Optional[int]
T3: TypeAlias = Optional[Any]


class TypingTest(TestCase):
    def test_assumptions(self):
        # self.assertTrue(isinstance(Any, type))
        self.assertTrue(isinstance(UnionType, type))
        self.assertTrue(not isinstance(Union, type))
        self.assertTrue(not isinstance(Union, UnionType))
        self.assertTrue(Union != UnionType)

        self.assertEqual(None, get_origin("NoTypesContainer"))
        self.assertEqual(None, get_origin("dict"))
        self.assertEqual(dict, get_origin(dict[str, "NoTypesContainer"]))
        self.assertEqual(
            (str, "NoTypesContainer"), get_args(dict[str, "NoTypesContainer"])
        )

        self.assertEqual(Union, get_origin(T1))
        self.assertEqual({bool, int, str, NoneType}, set(get_args(T1)))

        self.assertEqual(Union, get_origin(T2))
        self.assertEqual({int, NoneType}, set(get_args(T2)))

        self.assertEqual(Union, get_origin(T3))
        self.assertEqual({Any, NoneType}, set(get_args(T3)))


# noinspection PyMethodMayBeStatic
class ValueConstructibleTest(TestCase):
    def test_useless_ok(self):
        container = UselessContainer()
        self.assertIs(container, UselessContainer.from_value(container))

    # noinspection PyMethodMayBeStatic
    def test_useless_fail(self):
        with pytest.raises(
            TypeError,
            match="uc must be of type UselessContainer, but got None",
        ):
            UselessContainer.from_value(None, value_name="uc")

        with pytest.raises(
            TypeError,
            match="uc must be of type UselessContainer, but got bool",
        ):
            UselessContainer.from_value(True, value_name="uc")

        with pytest.raises(
            TypeError,
            match="uc must be of type UselessContainer, but got int",
        ):
            UselessContainer.from_value(1, value_name="uc")

        with pytest.raises(
            TypeError,
            match="uc must be of type UselessContainer, but got float",
        ):
            UselessContainer.from_value(0.1, value_name="uc")

        with pytest.raises(
            TypeError,
            match="uc must be of type UselessContainer, but got str",
        ):
            UselessContainer.from_value("abc", value_name="uc")

        with pytest.raises(
            TypeError,
            match="utc must be of type UselessContainer, but got dict",
        ):
            UselessContainer.from_value({}, value_name="utc")

        with pytest.raises(
            TypeError,
            match="uc must be of type UselessContainer, but got list",
        ):
            UselessContainer.from_value([], value_name="uc")

        with pytest.raises(
            TypeError,
            match="utc must be of type UselessContainer, but got object",
        ):
            UselessContainer.from_value(object(), value_name="utc")

        with pytest.raises(
            TypeError,
            match="utc must be of type UselessContainer, but got type",
        ):
            UselessContainer.from_value(int, value_name="utc")

        with pytest.raises(
            TypeError,
            match="utc must be of type UselessContainer, but got type",
        ):
            UselessContainer.from_value(UselessContainer, value_name="utc")

    def test_required_props_ok(self):
        rpc = RequiredPropsContainer.from_value({"x": 12.0, "y": 23.0, "z": 34.0})
        self.assertEqual(RequiredPropsContainer(x=12.0, y=23.0, z=34.0), rpc)

    # noinspection PyMethodMayBeStatic
    def test_required_props_fail(self):
        with pytest.raises(
            TypeError,
            match=(
                r"missing value for required property rpc.y"
                r" of type RequiredPropsContainer \| dict\[str, Any\]"
            ),
        ):
            RequiredPropsContainer.from_value({"x": 12.0, "z": 34.0}, "rpc")

    def test_no_types_ok(self):
        ntc = NoTypesContainer.from_value(dict(u=True, v=654, w="abc"))
        self.assertEqual(True, ntc.u)
        self.assertEqual(654, ntc.v)
        self.assertEqual("abc", ntc.w)


class MappingConstructibleTest(TestCase):
    def test_simple_ok(self):
        kwargs = dict(a="?", b=True, c=12, d=34.56, e="uvw", f=bytes, g="on")
        container = SimpleTypesContainer(**kwargs)
        self.assertEqual(container, SimpleTypesContainer.from_value(kwargs))
        self.assertIs(container, SimpleTypesContainer.from_value(container))

    def test_complex_ok(self):
        kwargs = {
            "p": {"a": None, "b": False, "c": 0, "d": 0.0, "e": "abc"},
            "q": {"p": True, "q": False},
            "r": {
                "u": {"a": None, "b": False, "c": 0, "d": 0.0, "e": "abc"},
                "v": {"a": None, "b": False, "c": 0, "d": 0.0, "e": "abc"},
            },
            "s": [1, 2, 3],
            "t": [
                {"a": None, "b": False, "c": 5, "d": 6.7, "e": "abc"},
                {"a": None, "b": False, "c": 8, "d": 9.1, "e": "abc", "f": str},
            ],
        }
        expected_container = ComplexTypesContainer(
            p=SimpleTypesContainer(a=None, b=False, c=0, d=0.0, e="abc"),
            q={"p": True, "q": False},
            r={
                "u": SimpleTypesContainer(a=None, b=False, c=0, d=0.0, e="abc"),
                "v": SimpleTypesContainer(a=None, b=False, c=0, d=0.0, e="abc"),
            },
            s=[1, 2, 3],
            t=[
                SimpleTypesContainer(a=None, b=False, c=5, d=6.7, e="abc"),
                SimpleTypesContainer(a=None, b=False, c=8, d=9.1, e="abc", f=str),
            ],
            u=None,
        )
        self.assertEqual(expected_container, ComplexTypesContainer.from_value(kwargs))
        self.assertIs(
            expected_container, ComplexTypesContainer.from_value(expected_container)
        )

    def test_union_ok(self):
        expected_union = UnionTypesContainer(m=SimpleTypesContainer())
        self.assertEqual(
            expected_union,
            UnionTypesContainer.from_value({"m": SimpleTypesContainer()}),
        )
        self.assertIs(expected_union, UnionTypesContainer.from_value(expected_union))

        expected_union = UnionTypesContainer(m=ComplexTypesContainer())
        self.assertEqual(
            expected_union,
            UnionTypesContainer.from_value({"m": ComplexTypesContainer()}),
        )
        self.assertIs(expected_union, UnionTypesContainer.from_value(expected_union))

        expected_union = UnionTypesContainer(m=None)
        self.assertEqual(
            expected_union,
            UnionTypesContainer.from_value({"m": None}),
        )
        self.assertIs(expected_union, UnionTypesContainer.from_value(expected_union))

    # noinspection PyMethodMayBeStatic
    def test_simple_fail(self):
        with pytest.raises(
            TypeError,
            match=(
                r"stc.b must be of type SimpleTypesContainer | dict\[str, Any\],"
                r" but got None"
            ),
        ):
            SimpleTypesContainer.from_value({"b": None}, value_name="stc")

        with pytest.raises(
            TypeError,
            match=r"stc.g must be one of 0, 1, True, False, 'on', 'off', but got 74",
        ):
            SimpleTypesContainer.from_value({"g": 74}, value_name="stc")

        with pytest.raises(
            TypeError, match="x is not a member of stc of type SimpleTypesContainer"
        ):
            SimpleTypesContainer.from_value({"x": 12}, value_name="stc")

        with pytest.raises(
            TypeError, match="x, y are not members of stc of type SimpleTypesContainer"
        ):
            SimpleTypesContainer.from_value({"x": 12, "y": 34}, value_name="stc")

        with pytest.raises(
            TypeError,
            match=(
                "mappings used to instantiate stc of type SimpleTypesContainer"
                " must have keys of type str, but found key of type int"
            ),
        ):
            SimpleTypesContainer.from_value({12: "x"}, value_name="stc")

        with pytest.raises(
            TypeError,
            match=(
                r"stc must be of type SimpleTypesContainer | dict\[str, Any\],"
                r" but got type"
            ),
        ):
            SimpleTypesContainer.from_value(SimpleTypesContainer, value_name="stc")

        with pytest.raises(
            TypeError,
            match="stc.f must be of type type, but got str",
        ):
            SimpleTypesContainer.from_value({"f": "pippo"}, value_name="stc")

    # noinspection PyMethodMayBeStatic
    def test_complex_fail(self):
        with pytest.raises(
            TypeError,
            match="keys of ctc.q must be of type str, but got bool",
        ):
            ComplexTypesContainer.from_value({"q": {True: False}}, value_name="ctc")

        with pytest.raises(
            TypeError,
            match=r"ctc.q\['x'\] must be of type bool, but got float",
        ):
            ComplexTypesContainer.from_value({"q": {"x": 2.3}}, value_name="ctc")

        with pytest.raises(
            TypeError,
            match=r"ctc.s\[1\] must be of type int, but got str",
        ):
            ComplexTypesContainer.from_value({"s": [1, "x", 3]}, value_name="ctc")

    # noinspection PyMethodMayBeStatic
    def test_union_fail(self):
        with pytest.raises(
            TypeError,
            match=(
                r"utc must be of type UnionTypesContainer | dict\[str, Any\],"
                r" but got int"
            ),
        ):
            UnionTypesContainer.from_value(21, value_name="utc")

        with pytest.raises(
            TypeError,
            match=(
                "utc.m must be of type SimpleTypesContainer"
                " | ComplexTypesContainer"
                " | None,"
                " but got str"
            ),
        ):
            UnionTypesContainer.from_value({"m": "pippo"}, value_name="utc")

    def test_get_class_parameters_is_cached(self):
        ctc_param = ComplexTypesContainer.class_parameters()
        stc_param = SimpleTypesContainer.class_parameters()
        self.assertIs(stc_param, SimpleTypesContainer.class_parameters())
        self.assertIs(ctc_param, ComplexTypesContainer.class_parameters())
        self.assertIsNot(ctc_param, stc_param)


class GetClassParametersTest(TestCase):
    def test_resolves_types(self):
        ctc_params = get_class_parameters(
            UnresolvedTypesContainer,
            forward_refs=UnresolvedTypesContainer.forward_refs(),
        )
        # order is important!
        self.assertEqual(
            [
                "a",
                "b",
                "c",
                "d",
                "e",
                "f",
                "g",
                "p",
                "q",
                "r",
                "s",
                "t",
                "u",
                "rules",
                "plugins",
            ],
            list(ctc_params.keys()),
        )
        for k, v in ctc_params.items():
            self.assertIsNotNone(v, msg=f"ctc_params[{k!r}]")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.formatters.json import Json

from .helpers import get_context, get_test_results


class JsonTest(TestCase):
    def test_json(self):
        results = get_test_results()
        formatter = Json()
        text = formatter.format(
            context=get_context(),
            results=results,
        )
        self.assertIn('"results": [', text)

    def test_json_with_meta(self):
        results = get_test_results()
        formatter = Json(with_meta=True)
        text = formatter.format(
            context=get_context(),
            results=results,
        )
        self.assertIn('"results": [', text)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

import xarray as xr

# noinspection PyProtectedMember
from xrlint._linter.rulectx import RuleContextImpl
from xrlint.config import ConfigObject
from xrlint.constants import DATASET_ROOT_NAME
from xrlint.result import Message, Suggestion


class RuleContextImplTest(TestCase):
    def test_defaults(self):
        config_obj = ConfigObject()
        dataset = xr.Dataset()
        context = RuleContextImpl(config_obj, dataset, "./ds.zarr", None, None)
        self.assertIs(config_obj, context.config)
        self.assertIs(dataset, context.dataset)
        self.assertEqual({}, context.settings)
        self.assertEqual("./ds.zarr", context.file_path)
        self.assertEqual(None, context.file_index)
        self.assertEqual(None, context.access_latency)

    def test_report(self):
        context = RuleContextImpl(
            ConfigObject(), xr.Dataset(), "./ds.zarr", None, 1.2345
        )
        with context.use_state(rule_id="no-xxx"):
            context.report(
                "What the heck do you mean?",
                suggestions=[Suggestion("Never say XXX again.")],
            )
            context.report("You said it.", fatal=True)
        self.assertEqual(
            [
                Message(
                    message="What the heck do you mean?",
                    node_path=DATASET_ROOT_NAME,
                    rule_id="no-xxx",
                    severity=2,
                    suggestions=[
                        Suggestion(desc="Never say XXX again.", data=None, fix=None)
                    ],
                ),
                Message(
                    message="You said it.",
                    node_path=DATASET_ROOT_NAME,
                    rule_id="no-xxx",
                    severity=2,
                    fatal=True,
                    suggestions=None,
                ),
            ],
            context.messages,
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np

from xrlint.node import VariableNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "var-missing-data",
    version="1.0.0",
    type="suggestion",
    description=(
        "Checks the recommended use of missing data, i.e., coordinate variables"
        " should not define missing data, but packed data should."
        " Notifies about the use of valid ranges to indicate missing data, which"
        " is currently not supported by xarray."
    ),
    docs_url="https://cfconventions.org/cf-conventions/cf-conventions.html#units",
)
class VarMissingData(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        array = node.array
        encoding = array.encoding
        attrs = array.attrs

        fill_value_source = None
        if "_FillValue" in encoding:
            fill_value_source = "encoding"
        elif "_FillValue" in attrs:
            fill_value_source = "attribute"

        if fill_value_source is not None and node.name in ctx.dataset.coords:
            ctx.report(
                f"Unexpected {fill_value_source} '_FillValue',"
                f" coordinates must not have missing data."
            )
        elif fill_value_source is None and node.name in ctx.dataset.data_vars:
            scaling_factor = encoding.get("scaling_factor", attrs.get("scaling_factor"))
            add_offset = encoding.get("add_offset", attrs.get("add_offset"))
            raw_dtype = encoding.get("dtype")
            if add_offset is not None or scaling_factor is not None:
                ctx.report("Missing attribute '_FillValue' since data packing is used.")
            elif isinstance(raw_dtype, np.dtype) and np.issubdtype(
                raw_dtype, np.floating
            ):
                ctx.report("Missing attribute '_FillValue', which should be NaN.")

        if any((name in attrs) for name in ("valid_min", "valid_max", "valid_range")):
            ctx.report("Valid ranges are not recognized by xarray (as of Feb 2025).")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Final

LON_NAME: Final = "lon"
LAT_NAME: Final = "lat"
X_NAME: Final = "x"
Y_NAME: Final = "y"
TIME_NAME: Final = "time"

GM_NAMES: Final = "spatial_ref", "crs"
GM_NAMES_TEXT: Final = " or ".join(repr(gm_name) for gm_name in GM_NAMES)

ML_FILE_PATTERN: Final = "**/*.levels"
ML_META_FILENAME: Final = ".zlevels"
ML_INFO_ATTR: Final = "_LEVEL_INFO"



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.core.rules.time_coordinate import TimeCoordinate
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset()
valid_dataset_1 = xr.Dataset(
    coords={
        "time": xr.DataArray(
            np.array([3, 4, 5], dtype=np.dtype("datetime64[s]")),
            dims="time",
            attrs={
                "standard_name": "time",
            },
        ),
    },
    data_vars={
        "pos": xr.DataArray([10, 20, 30], dims="time", attrs={"units": "seconds"})
    },
)
valid_dataset_1.time.encoding["units"] = "seconds since 2000-01-01 00:00:00 +2:00"
valid_dataset_1.time.encoding["calendar"] = "gregorian"

# OK, because with decode_cf=False meta-info is in attrs still
valid_dataset_2 = valid_dataset_1.copy()
del valid_dataset_2.time.encoding["units"]
del valid_dataset_2.time.encoding["calendar"]
valid_dataset_2.time.attrs["units"] = "seconds since 2000-1-1 +2:00"
valid_dataset_2.time.attrs["calendar"] = "gregorian"

# OK, because not identified as time
valid_dataset_3 = valid_dataset_1.copy()
del valid_dataset_3.time.encoding["units"]
del valid_dataset_3.time.attrs["standard_name"]

# OK, because we only look for time units
valid_dataset_4 = valid_dataset_1.rename_vars({"time": "tm"})
del valid_dataset_4.tm.attrs["standard_name"]

# OK, because not recognized as time coord
valid_dataset_5 = valid_dataset_1.copy()
valid_dataset_5.time.encoding["units"] = 1
del valid_dataset_5.time.attrs["standard_name"]

# Invalid, because units is invalid but standard_name given
invalid_dataset_0 = valid_dataset_1.copy()
invalid_dataset_0.time.encoding["units"] = 1

# Invalid, because we require units but standard_name given
invalid_dataset_1 = valid_dataset_1.copy(deep=True)
del invalid_dataset_1.time.encoding["units"]

# Invalid, because we no time units although standard_name given
invalid_dataset_2 = valid_dataset_1.copy(deep=True)
invalid_dataset_2.time.encoding["units"] = "years from 2000-1-1 +0:0"

# Invalid, because we require calendar
invalid_dataset_3 = valid_dataset_1.copy(deep=True)
del invalid_dataset_3.time.encoding["calendar"]

# Invalid, because we use invalid UOT
invalid_dataset_4 = valid_dataset_1.copy(deep=True)
invalid_dataset_4.time.encoding["units"] = "millis since 2000-1-1 +0:0"

# Invalid, because we use ambiguous UOT
invalid_dataset_5 = valid_dataset_1.copy(deep=True)
invalid_dataset_5.time.encoding["units"] = "years since 2000-1-1 +0:0"

# Invalid, because we require timezone
invalid_dataset_6 = valid_dataset_1.copy(deep=True)
invalid_dataset_6.time.encoding["units"] = "seconds since 2000-01-01 00:00:00"

# Invalid, because we require timezone
invalid_dataset_7 = valid_dataset_1.copy(deep=True)
invalid_dataset_7.time.encoding["units"] = "seconds since 2000-01-01"

# Invalid, because we have 6 units parts
invalid_dataset_8 = valid_dataset_1.copy(deep=True)
invalid_dataset_8.time.encoding["units"] = "days since 2000-01-01 12:00:00 +0:00 utc"

# Invalid, because we date part is invalid
invalid_dataset_9 = valid_dataset_1.copy(deep=True)
invalid_dataset_9.time.encoding["units"] = "days since 00-01-01 12:00:00 +0:00"

# Invalid, because we time part is invalid
invalid_dataset_10 = valid_dataset_1.copy(deep=True)
invalid_dataset_10.time.encoding["units"] = "days since 2000-01-01 12:00 +0:00"

# Invalid, because we tz part is invalid
invalid_dataset_11 = valid_dataset_1.copy(deep=True)
invalid_dataset_11.time.encoding["units"] = "days since 2000-01-01 12:00:00 utc"

TimeCoordinateTest = RuleTester.define_test(
    "time-coordinate",
    TimeCoordinate,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
        RuleTest(dataset=valid_dataset_4),
        RuleTest(dataset=valid_dataset_5),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=1),
        RuleTest(dataset=invalid_dataset_1, expected=1),
        RuleTest(dataset=invalid_dataset_2, expected=1),
        RuleTest(dataset=invalid_dataset_3, expected=1),
        RuleTest(dataset=invalid_dataset_4, expected=1),
        RuleTest(dataset=invalid_dataset_5, expected=1),
        RuleTest(dataset=invalid_dataset_6, expected=1),
        RuleTest(dataset=invalid_dataset_7, expected=1),
        RuleTest(dataset=invalid_dataset_8, expected=1),
        RuleTest(dataset=invalid_dataset_9, expected=1),
        RuleTest(dataset=invalid_dataset_10, expected=1),
        RuleTest(dataset=invalid_dataset_11, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import Any, Callable, Type

from xrlint.operation import Operation, OperationMeta
from xrlint.result import Result, ResultStats


class FormatterContext(ABC):
    """A formatter context is passed to `FormatOp`."""

    @property
    @abstractmethod
    def max_warnings_exceeded(self) -> bool:
        """`True` if the maximum number of warnings has been exceeded."""

    @property
    @abstractmethod
    def result_stats(self) -> ResultStats:
        """Get current result statistics."""


class FormatterOp(ABC):
    """Define the specific format operation."""

    @abstractmethod
    def format(
        self,
        context: FormatterContext,
        results: Iterable[Result],
    ) -> str:
        """Format the given results.

        Args:
            context: formatting context
            results: an iterable of results to format
        Returns:
            A text representing the results in a given format
        """


@dataclass(kw_only=True)
class FormatterMeta(OperationMeta):
    """Formatter metadata."""

    name: str
    """Formatter name."""

    version: str = "0.0.0"
    """Formatter version."""

    ref: str | None = None
    """Formatter reference.
    Specifies the location from where the formatter can be
    dynamically imported.
    Must have the form "<module>:<attr>", if given.
    """

    schema: dict[str, Any] | list[dict[str, Any]] | bool | None = None
    """Formatter options schema."""


@dataclass(frozen=True, kw_only=True)
class Formatter(Operation):
    """A formatter for linting results."""

    meta: FormatterMeta
    """The formatter metadata."""

    op_class: Type[FormatterOp]
    """The class that implements the format operation."""

    @classmethod
    def meta_class(cls) -> Type:
        return FormatterMeta

    @classmethod
    def op_base_class(cls) -> Type:
        return FormatterOp

    @classmethod
    def value_name(cls) -> str:
        return "formatter"


class FormatterRegistry(Mapping[str, Formatter]):
    def __init__(self):
        self._registrations = {}

    def define_formatter(
        self,
        name: str | None = None,
        version: str | None = None,
        schema: dict[str, Any] | list[dict[str, Any]] | bool | None = None,
    ) -> Callable[[FormatterOp], Type[FormatterOp]] | Formatter:
        """Decorator function."""
        return Formatter.define_operation(
            None,
            registry=self._registrations,
            meta_kwargs=dict(name=name, version=version, schema=schema),
        )

    def __getitem__(self, key: str) -> Formatter:
        return self._registrations[key]

    def __len__(self) -> int:
        return len(self._registrations)

    def __iter__(self):
        return iter(self._registrations)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.plugins.xcube import export_plugin


class ExportPluginTest(TestCase):
    def test_rules_complete(self):
        plugin = export_plugin()
        self.assertEqual(
            {
                "any-spatial-data-var",
                "cube-dims-order",
                "data-var-colors",
                "dataset-title",
                "grid-mapping-naming",
                "increasing-time",
                "lat-lon-naming",
                "ml-dataset-meta",
                "ml-dataset-time",
                "ml-dataset-xy",
                "no-chunked-coords",
                "single-grid-mapping",
                "time-naming",
            },
            set(plugin.rules.keys()),
        )

    def test_configs_complete(self):
        plugin = export_plugin()
        self.assertEqual(
            {
                "all",
                "recommended",
            },
            set(plugin.configs.keys()),
        )
        all_rule_names = set(f"xcube/{k}" for k in plugin.rules.keys())
        self.assertEqual(
            all_rule_names,
            set(plugin.configs["all"][-1].rules.keys()),
        )
        self.assertEqual(
            all_rule_names,
            set(plugin.configs["recommended"][-1].rules.keys()),
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.util.formatting import (
    format_case,
    format_count,
    format_problems,
    format_seq,
    format_styled,
)


class FormattingTest(TestCase):
    def test_format_count(self):
        self.assertEqual("-3 eggs", format_count(-3, "egg"))
        self.assertEqual("-2 eggs", format_count(-2, "egg"))
        self.assertEqual("-1 eggs", format_count(-1, "egg"))
        self.assertEqual("no eggs", format_count(0, "egg"))
        self.assertEqual("one egg", format_count(1, "egg"))
        self.assertEqual("2 eggs", format_count(2, "egg"))
        self.assertEqual("3 eggs", format_count(3, "egg"))

    def test_format_problems(self):
        self.assertEqual("no problems", format_problems(0, 0))
        self.assertEqual("one error", format_problems(1, 0))
        self.assertEqual("2 errors", format_problems(2, 0))
        self.assertEqual("one warning", format_problems(0, 1))
        self.assertEqual(
            "2 problems (one error and one warning)", format_problems(1, 1)
        )
        self.assertEqual("3 problems (2 errors and one warning)", format_problems(2, 1))
        self.assertEqual("2 warnings", format_problems(0, 2))
        self.assertEqual("3 problems (one error and 2 warnings)", format_problems(1, 2))
        self.assertEqual("4 problems (2 errors and 2 warnings)", format_problems(2, 2))

    def test_format_case(self):
        self.assertEqual("hello", format_case("hello"))
        self.assertEqual("Hello", format_case("Hello"))
        self.assertEqual("hello", format_case("hello", upper=False))
        self.assertEqual("Hello", format_case("hello", upper=True))
        self.assertEqual("hello", format_case("Hello", upper=False))
        self.assertEqual("Hello", format_case("Hello", upper=True))

    def test_format_seq(self):
        self.assertEqual("", format_seq([]))
        self.assertEqual("1, 2, 3", format_seq([1, 2, 3]))
        self.assertEqual(
            "1, 2, 3, ..., 8, 9, 10", format_seq([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        )
        self.assertEqual(
            "1, 2, ..., 9, 10", format_seq([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], max_count=4)
        )

    def test_format_styled(self):
        self.assertEqual("Hello", format_styled("Hello"))
        self.assertEqual("\x1b[2mHello\x1b[0m", format_styled("Hello", s="dim"))
        self.assertEqual("\x1b[;31mHello\x1b[0m", format_styled("Hello", fg="red"))
        self.assertEqual("\x1b[;;42mHello\x1b[0m", format_styled("Hello", bg="green"))
        self.assertEqual(
            "\x1b]8;;file://x.nc\x1b\\File\x1b]8;;\x1b\\",
            format_styled("File", href="x.nc"),
        )
        self.assertEqual(
            "\x1b]8;;https://data.com/x.nc\x1b\\File\x1b]8;;\x1b\\",
            format_styled("File", href="https://data.com/x.nc"),
        )
        self.assertEqual(
            "\x1b]8;;file://x.nc\x1b\\x.nc\x1b]8;;\x1b\\",
            format_styled("", href="x.nc"),
        )
        self.assertEqual(
            "\x1b]8;;file://x.nc\x1b\\\x1b[4mFile\x1b[0m\x1b]8;;\x1b\\",
            format_styled("File", href="x.nc", s="underline"),
        )

        self.assertEqual("", format_styled(""))
        self.assertEqual("", format_styled("", s="dim"))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from tests.plugins.xcube.helpers import make_cube_levels
from xrlint.plugins.xcube.rules.ml_dataset_time import MLDatasetTime
from xrlint.plugins.xcube.util import LevelsMeta
from xrlint.testing import RuleTest, RuleTester

meta = LevelsMeta(
    version="1.0",
    num_levels=4,
    use_saved_levels=True,
    agg_methods={"chl": "mean"},
)
levels_with_time = make_cube_levels(4, 720, 360, nt=6, meta=meta)
levels_wo_time = make_cube_levels(4, 720, 360, meta=meta)

valid_dataset_0 = levels_with_time[0]
valid_dataset_1 = levels_with_time[1]
valid_dataset_2 = levels_wo_time[0]
valid_dataset_3 = xr.Dataset()

invalid_dataset_0 = levels_with_time[0].copy()
invalid_dataset_0["chl"] = invalid_dataset_0["chl"].chunk(time=3)

MLDatasetTimeTest = RuleTester.define_test(
    "ml-dataset-time",
    MLDatasetTime,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.xcube.rules.cube_dims_order import CubeDimsOrder
from xrlint.testing import RuleTest, RuleTester


def make_dataset(dims: tuple[str, str, str]):
    n = 3
    return xr.Dataset(
        attrs=dict(title="v-data"),
        coords={
            "x": xr.DataArray(np.linspace(0, 1, n), dims="x", attrs={"units": "m"}),
            "y": xr.DataArray(np.linspace(0, 1, n), dims="y", attrs={"units": "m"}),
            "time": xr.DataArray(
                list(range(2010, 2010 + n)), dims="time", attrs={"units": "years"}
            ),
        },
        data_vars={
            "chl": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
            "tsm": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
            "avg_temp": xr.DataArray(
                np.random.random(n), dims=dims[0], attrs={"units": "kelvin"}
            ),
        },
    )


valid_dataset_0 = make_dataset(("time", "y", "x"))
valid_dataset_1 = make_dataset(("time", "lat", "lon"))
valid_dataset_2 = make_dataset(("level", "y", "x"))

invalid_dataset_0 = make_dataset(("time", "x", "y"))
invalid_dataset_1 = make_dataset(("x", "y", "time"))
invalid_dataset_2 = make_dataset(("time", "lon", "lat"))
invalid_dataset_3 = make_dataset(("lon", "lat", "level"))
invalid_dataset_4 = make_dataset(("x", "y", "level"))


CubeDimsOrderTest = RuleTester.define_test(
    "cube-dims-order",
    CubeDimsOrder,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=2),
        RuleTest(dataset=invalid_dataset_1, expected=2),
        RuleTest(dataset=invalid_dataset_2, expected=2),
        RuleTest(dataset=invalid_dataset_3, expected=2),
        RuleTest(dataset=invalid_dataset_4, expected=2),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from collections.abc import Iterable

from tabulate import tabulate

from xrlint.constants import SEVERITY_CODE_TO_COLOR, SEVERITY_CODE_TO_NAME
from xrlint.formatter import FormatterContext, FormatterOp
from xrlint.formatters import registry
from xrlint.result import Message, Result
from xrlint.util.formatting import format_problems, format_styled
from xrlint.util.schema import schema


@registry.define_formatter(
    "simple",
    version="1.0.0",
    schema=schema(
        "object",
        properties=dict(
            styled=schema("boolean", default=True),
            output=schema("boolean", default=True),
        ),
    ),
)
class Simple(FormatterOp):
    """Simple output formatter.
    Produces either ANSI-styled (default) or plain text reports.
    It incrementally outputs results to console (stdout) by default.
    """

    def __init__(self, styled: bool = True, output: bool = True):
        self.styled = styled
        self.output = output

    def format(
        self,
        context: FormatterContext,
        results: Iterable[Result],
    ) -> str:
        text_parts = []

        error_count = 0
        warning_count = 0
        for result in results:
            result_text = self.format_result(result)
            if self.output:
                print(result_text, flush=True, end="")
            text_parts.append(result_text)
            error_count += result.error_count
            warning_count += result.warning_count

        summary_text = self._format_summary(error_count, warning_count)
        if self.output:
            print(summary_text, flush=True, end="")
        text_parts.append(summary_text)

        return "".join(text_parts)

    def format_result(
        self,
        result: Result,
    ) -> str:
        file_path_text = self._format_file_path(result)
        if not result.messages:
            return f"\n{file_path_text} - ok\n"
        result_parts = [f"\n{file_path_text}:\n"]
        result_data = []
        for message in result.messages:
            result_data.append(
                [
                    self._format_node_path(message),
                    self._format_severity(message),
                    self._format_message(message),
                    self._format_rule_id(message, result),
                ]
            )
        result_parts.append(tabulate(result_data, headers=(), tablefmt="plain"))
        result_parts.append("\n")
        return "".join(result_parts)

    def _format_file_path(self, result) -> str:
        file_path_text = result.file_path
        if self.styled:
            file_path_text = format_styled(file_path_text, s="underline")
        return file_path_text

    def _format_node_path(self, m: Message) -> str:
        node_text = m.node_path or ""
        if self.styled and node_text:
            node_text = format_styled(node_text, s="dim")
        return node_text

    def _format_severity(self, m: Message) -> str:
        severity_text = SEVERITY_CODE_TO_NAME.get(m.severity, "?")
        if self.styled and severity_text:
            fg = SEVERITY_CODE_TO_COLOR.get(m.severity, "")
            severity_text = format_styled(severity_text, s="bold", fg=fg)
        return severity_text

    # noinspection PyMethodMayBeStatic
    def _format_message(self, m: Message) -> str:
        return m.message or ""

    def _format_rule_id(self, m: Message, r: Result) -> str:
        rule_text = m.rule_id or ""
        if self.styled and rule_text:
            rule_url = r.get_docs_url_for_rule(m.rule_id)
            if rule_url:
                rule_text = format_styled(rule_text, fg="blue", href=rule_url)
        return rule_text

    def _format_summary(self, error_count, warning_count) -> str:
        summary_parts = []
        problems_text = format_problems(error_count, warning_count)
        if self.styled:
            if error_count:
                problems_text = format_styled(
                    problems_text, fg=SEVERITY_CODE_TO_COLOR[2]
                )
            elif warning_count:
                problems_text = format_styled(
                    problems_text, fg=SEVERITY_CODE_TO_COLOR[1]
                )
        summary_parts.append("\n")
        summary_parts.append(problems_text)
        summary_parts.append("\n\n")
        return "".join(summary_parts)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any
from unittest import TestCase

import pytest
import xarray as xr

from xrlint.config import Config, ConfigObject, get_core_config_object
from xrlint.constants import CORE_PLUGIN_NAME
from xrlint.plugin import Plugin, new_plugin
from xrlint.processor import ProcessorOp, define_processor
from xrlint.result import Message
from xrlint.rule import Rule, RuleConfig
from xrlint.util.filefilter import FileFilter


# noinspection PyMethodMayBeStatic
class ConfigObjectTest(TestCase):
    def test_class_props(self):
        self.assertEqual("config_obj", ConfigObject.value_name())
        self.assertEqual("ConfigObject | dict | None", ConfigObject.value_type_name())

    def test_defaults(self):
        config_obj = ConfigObject()
        self.assertEqual(None, config_obj.name)
        self.assertEqual(None, config_obj.files)
        self.assertEqual(None, config_obj.ignores)
        self.assertEqual(None, config_obj.linter_options)
        self.assertEqual(None, config_obj.opener_options)
        self.assertEqual(None, config_obj.processor)
        self.assertEqual(None, config_obj.plugins)
        self.assertEqual(None, config_obj.rules)

    def test_get_plugin(self):
        config_obj = get_core_config_object()
        plugin = config_obj.get_plugin(CORE_PLUGIN_NAME)
        self.assertIsInstance(plugin, Plugin)

        with pytest.raises(ValueError, match="unknown plugin 'xcube'"):
            config_obj.get_plugin("xcube")

    def test_get_rule(self):
        config_obj = get_core_config_object()
        rule = config_obj.get_rule("var-flags")
        self.assertIsInstance(rule, Rule)

        with pytest.raises(ValueError, match="unknown rule 'foo'"):
            config_obj.get_rule("foo")

    def test_get_processor_op(self):
        class MyProc(ProcessorOp):
            def preprocess(
                self, file_path: str, opener_options: dict[str, Any]
            ) -> list[tuple[xr.Dataset, str]]:
                pass

            def postprocess(
                self, messages: list[list[Message]], file_path: str
            ) -> list[Message]:
                pass

        processor = define_processor("myproc", op_class=MyProc)
        config_obj = ConfigObject(
            plugins=dict(
                myplugin=new_plugin("myplugin", processors=dict(myproc=processor))
            )
        )

        processor_op = config_obj.get_processor_op(MyProc())
        self.assertIsInstance(processor_op, MyProc)

        processor_op = config_obj.get_processor_op("myplugin/myproc")
        self.assertIsInstance(processor_op, MyProc)

        with pytest.raises(ValueError, match="unknown processor 'myplugin/myproc2'"):
            config_obj.get_processor_op("myplugin/myproc2")

    def test_from_value_ok(self):
        self.assertEqual(ConfigObject(), ConfigObject.from_value(None))
        self.assertEqual(ConfigObject(), ConfigObject.from_value({}))
        self.assertEqual(ConfigObject(), ConfigObject.from_value(ConfigObject()))
        self.assertEqual(
            ConfigObject(name="x"), ConfigObject.from_value(ConfigObject(name="x"))
        )
        self.assertEqual(
            ConfigObject(
                name="xXx",
                files=["**/*.zarr", "**/*.nc"],
                linter_options={"a": 4},
                opener_options={"b": 5},
                settings={"c": 6},
            ),
            ConfigObject.from_value(
                {
                    "name": "xXx",
                    "files": ["**/*.zarr", "**/*.nc"],
                    "linter_options": {"a": 4},  # not used yet
                    "opener_options": {"b": 5},  # not used yet
                    "settings": {"c": 6},
                }
            ),
        )
        self.assertEqual(
            ConfigObject(
                rules={
                    "hello/no-spaces-in-titles": RuleConfig(severity=2),
                    "hello/time-without-tz": RuleConfig(severity=0),
                    "hello/no-empty-units": RuleConfig(
                        severity=1, args=(12,), kwargs={"indent": 4}
                    ),
                },
            ),
            ConfigObject.from_value(
                {
                    "rules": {
                        "hello/no-spaces-in-titles": 2,
                        "hello/time-without-tz": "off",
                        "hello/no-empty-units": ["warn", 12, {"indent": 4}],
                    },
                }
            ),
        )

    def test_to_json(self):
        config_obj = ConfigObject(
            name="xXx",
            files=["**/*.zarr", "**/*.nc"],
            linter_options={"a": 4},
            opener_options={"b": 5},
            settings={"c": 6},
            rules={
                "hello/no-spaces-in-titles": RuleConfig(severity=2),
                "hello/time-without-tz": RuleConfig(severity=0),
                "hello/no-empty-units": RuleConfig(
                    severity=1, args=(12,), kwargs={"indent": 4}
                ),
            },
        )
        self.assertEqual(
            {
                "name": "xXx",
                "files": ["**/*.zarr", "**/*.nc"],
                "linter_options": {"a": 4},
                "opener_options": {"b": 5},
                "settings": {"c": 6},
                "rules": {
                    "hello/no-empty-units": [1, 12, {"indent": 4}],
                    "hello/no-spaces-in-titles": 2,
                    "hello/time-without-tz": 0,
                },
            },
            config_obj.to_json(),
        )

    def test_from_value_fails(self):
        with pytest.raises(
            TypeError,
            match=r"config_obj must be of type ConfigObject \| dict \| None, but got int",
        ):
            ConfigObject.from_value(4)

        with pytest.raises(
            TypeError,
            match=r"config_obj must be of type ConfigObject \| dict \| None, but got str",
        ):
            ConfigObject.from_value("abc")

        with pytest.raises(
            TypeError,
            match=r"config_obj must be of type ConfigObject \| dict \| None, but got tuple",
        ):
            ConfigObject.from_value(())

        with pytest.raises(
            TypeError,
            match=r" config_obj.linter_options must be of type dict.*, but got list",
        ):
            ConfigObject.from_value({"linter_options": [1, 2, 3]})

        with pytest.raises(
            TypeError,
            match=r" keys of config_obj.settings must be of type str, but got int",
        ):
            ConfigObject.from_value({"settings": {8: 9}})


class ConfigTest(TestCase):
    def test_from_config_ok(self):
        config = Config.from_config()
        self.assertIsInstance(config, Config)
        self.assertEqual([], config.objects)

        config = Config.from_config(
            {"ignores": ["**/*.levels"]},
            get_core_config_object(),
            "recommended",
            {"rules": {"no-empty-chunks": 2}},
        )
        self.assertIsInstance(config, Config)
        self.assertEqual(4, len(config.objects))

        config = Config.from_config(config)
        self.assertIsInstance(config, Config)
        self.assertEqual(4, len(config.objects))

        config = Config.from_config(config.objects)
        self.assertIsInstance(config, Config)
        self.assertEqual(4, len(config.objects))

        config = Config.from_config(*config.objects)
        self.assertIsInstance(config, Config)
        self.assertEqual(4, len(config.objects))

    def test_from_value_ok(self):
        config = Config.from_value([])
        self.assertIsInstance(config, Config)
        self.assertEqual([], config.objects)

        config_2 = Config.from_value(config)
        self.assertIs(config_2, config)

        config = Config.from_value([{}])
        self.assertIsInstance(config, Config)
        self.assertEqual([], config.objects)

        config_object = ConfigObject.from_value({})
        config = Config.from_value(config_object)
        self.assertIsInstance(config, Config)
        self.assertIs(config_object, config.objects[0])

    # noinspection PyMethodMayBeStatic
    def test_from_value_fail(self):
        with pytest.raises(
            TypeError,
            match=(
                r"config must be of type"
                r" Config \| ConfigObjectLike \| str \| Sequence\[ConfigObjectLike \| str\],"
                r" but got int"
            ),
        ):
            Config.from_value(264)

    def test_compute_config(self):
        config = Config([ConfigObject()])
        file_path = "s3://wq-services/datacubes/chl-2.zarr"
        self.assertEqual(ConfigObject(), config.compute_config_object(file_path))

        config = Config(
            [
                ConfigObject(ignores=["**/*.yaml"], settings={"a": 1, "b": 1}),
                ConfigObject(files=["**/datacubes/*.zarr"], settings={"b": 2}),
                ConfigObject(files=["**/*.txt"], settings={"a": 2}),
            ]
        )
        file_path = "s3://wq-services/datacubes/chl-2.zarr"
        self.assertEqual(
            ConfigObject(settings={"a": 1, "b": 2}),
            config.compute_config_object(file_path),
        )

        # global ignores
        file_path = "s3://wq-services/datacubes/chl-2.txt"
        self.assertEqual(
            ConfigObject(settings={"a": 2, "b": 1}),
            config.compute_config_object(file_path),
        )

        file_path = "s3://wq-services/datacubes/config.yaml"
        self.assertEqual(
            None,
            config.compute_config_object(file_path),
        )

    def test_split_global_filter(self):
        config = Config(
            [
                ConfigObject(files=["**/*.hdf"]),  # global file
                ConfigObject(ignores=["**/chl-?.txt"]),  # global ignores
                ConfigObject(ignores=["**/chl-?.*"], settings={"a": 2}),
                ConfigObject(settings={"a": 1, "b": 1}),
                ConfigObject(files=["**/datacubes/*.zarr"], settings={"b": 2}),
            ]
        )

        new_config, file_filter = config.split_global_filter()
        self.assertEqual(
            FileFilter.from_patterns(["**/*.hdf"], ["**/chl-?.txt"]),
            file_filter,
        )
        self.assertEqual(3, len(new_config.objects))

        new_config, file_filter = config.split_global_filter(
            default=FileFilter.from_patterns(["**/*.h5"], None)
        )
        self.assertEqual(
            FileFilter.from_patterns(["**/*.h5", "**/*.hdf"], ["**/chl-?.txt"]),
            file_filter,
        )
        self.assertEqual(3, len(new_config.objects))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import math

import numpy as np
import xarray as xr

from xrlint.plugins.xcube.util import LevelsMeta, attach_dataset_level_infos


def make_cube_levels(
    nl: int,
    nx: int,
    ny: int,
    nt: int | None = None,
    meta: LevelsMeta | None = None,
    force_infos: bool = False,
) -> list[xr.Dataset]:
    levels = [
        make_cube(math.ceil(nx >> level), math.ceil(ny >> level), nt)
        for level in range(nl)
    ]
    if meta is not None or force_infos:
        attach_dataset_level_infos(
            [(ds, f"{i}.zarr") for i, ds in enumerate(levels)], meta=meta
        )
    return levels


def make_cube(nx: int, ny: int, nt: int | None = None) -> xr.Dataset:
    """Make an in-memory dataset that should pass all xcube rules.

    Args:
        nx: length of the lon-dimension
        ny: length of the lat-dimension
        nt: length of the time-dimension, optional

    Returns:
        an in-memory dataset with one 3-d data variable "chl"
            with dimensions ["time",] "lat", "lon".
    """
    x_attrs = dict(
        long_name="longitude",
        standard_name="longitude",
        units="degrees_east",
    )
    y_attrs = dict(
        long_name="latitude",
        standard_name="latitude",
        units="degrees_north",
    )

    dx = 180.0 / nx
    dy = 90.0 / ny
    x_data = np.linspace(-180 + dx, 180 - dx, nx)
    y_data = np.linspace(-90 + dy, 90 - dy, ny)

    chl_attrs = dict(
        long_name="chlorophyll concentration",
        standard_name="chlorophyll_concentration",
        units="mg/m^3",
        _FillValue=0,
    )
    chl_chunks = dict(lat=min(ny, 90), lon=min(nx, 90))

    ds_attrs = dict(title="Chlorophyll")

    coords = dict(
        lon=xr.DataArray(x_data, dims="lon", attrs=x_attrs),
        lat=xr.DataArray(y_data, dims="lat", attrs=y_attrs),
    )

    if nt is None:
        return xr.Dataset(
            data_vars=dict(
                chl=xr.DataArray(
                    np.zeros((ny, nx)), dims=["lat", "lon"], attrs=chl_attrs
                ).chunk(**chl_chunks),
            ),
            coords=coords,
            attrs=ds_attrs,
        )
    else:
        time_attrs = dict(
            long_name="time",
            standard_name="time",
            units="days since 2024-06-10:12:00:00 utc",
            calendar="gregorian",
        )
        coords.update(time=xr.DataArray(range(nt), dims="time", attrs=time_attrs))
        return xr.Dataset(
            data_vars=dict(
                chl=xr.DataArray(
                    np.zeros((nt, ny, nx)), dims=["time", "lat", "lon"], attrs=chl_attrs
                ).chunk(time=1, **chl_chunks),
            ),
            coords=coords,
            attrs=ds_attrs,
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase


class AllTest(TestCase):
    def test_api_is_complete(self):
        import xrlint.all as xrl
        from xrlint.all import __all__

        # noinspection PyUnresolvedReferences
        keys = set(
            k
            for k, v in xrl.__dict__.items()
            if isinstance(k, str) and not k.startswith("_")
        )
        self.assertEqual(set(__all__), keys)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from dataclasses import dataclass, field
from typing import Any, Literal
from unittest import TestCase

import pytest

from xrlint.util.serializable import JsonSerializable


class PlainSimpleTypesContainer(JsonSerializable):
    def __init__(
        self,
        a: Any = None,
        b: bool = False,
        c: int = 0,
        d: float = 0.0,
        e: str = "abc",
        f: type = int,
        g: Literal[0, 1, True, False, "on", "off"] = 0,
    ):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f
        self.g = g


class PlainComplexTypesContainer(JsonSerializable):
    def __init__(
        self,
        p: PlainSimpleTypesContainer = PlainSimpleTypesContainer(),
        q: dict[str, bool] = None,
        r: dict[str, PlainSimpleTypesContainer] = None,
        s: list[int] = None,
        t: list[PlainSimpleTypesContainer] = None,
        u: int | float | None = None,
    ):
        self.p = p
        self.q = q or {}
        self.r = r or {}
        self.s = s or []
        self.t = t or []
        self.u = u


@dataclass()
class DataclassSimpleTypesContainer(JsonSerializable):
    a: Any = None
    b: bool = False
    c: int = 0
    d: float = 0.0
    e: str = "abc"
    f: type = int
    g: Literal[0, 1, True, False, "on", "off"] = 0


@dataclass()
class DataclassComplexTypesContainer(JsonSerializable):
    p: DataclassSimpleTypesContainer = field(
        default_factory=DataclassSimpleTypesContainer
    )
    q: dict[str, bool] = field(default_factory=dict)
    r: dict[str, DataclassSimpleTypesContainer] = field(default_factory=dict)
    s: list[int] = field(default_factory=list)
    t: list[DataclassSimpleTypesContainer] = field(default_factory=list)
    u: int | float | None = None


# noinspection PyMethodMayBeStatic
class JsonSerializableTest(TestCase):
    def test_plain_simple_ok(self):
        self.assertEqual(
            {
                "a": None,
                "b": False,
                "c": 0,
                "d": 0.0,
                "e": "abc",
                "f": "<class 'int'>",
                "g": 0,
            },
            PlainSimpleTypesContainer().to_json(),
        )
        self.assertEqual(
            {
                "a": "?",
                "b": True,
                "c": 12,
                "d": 34.56,
                "e": "uvw",
                "f": "<class 'bool'>",
                "g": "off",
            },
            PlainSimpleTypesContainer(
                a="?", b=True, c=12, d=34.56, e="uvw", f=bool, g="off"
            ).to_json(),
        )

    def test_plain_complex_ok(self):
        container = PlainComplexTypesContainer(
            q=dict(p=True, q=False),
            r=dict(u=PlainSimpleTypesContainer(), v=PlainSimpleTypesContainer()),
            s=[1, 2, 3],
            t=[
                PlainSimpleTypesContainer(c=5, d=6.7),
                PlainSimpleTypesContainer(c=8, d=9.1, f=PlainSimpleTypesContainer),
            ],
        )
        self.assertEqual(
            {
                "p": {
                    "a": None,
                    "b": False,
                    "c": 0,
                    "d": 0.0,
                    "e": "abc",
                    "f": "<class 'int'>",
                    "g": 0,
                },
                "q": {"p": True, "q": False},
                "r": {
                    "u": {
                        "a": None,
                        "b": False,
                        "c": 0,
                        "d": 0.0,
                        "e": "abc",
                        "f": "<class 'int'>",
                        "g": 0,
                    },
                    "v": {
                        "a": None,
                        "b": False,
                        "c": 0,
                        "d": 0.0,
                        "e": "abc",
                        "f": "<class 'int'>",
                        "g": 0,
                    },
                },
                "s": [1, 2, 3],
                "t": [
                    {
                        "a": None,
                        "b": False,
                        "c": 5,
                        "d": 6.7,
                        "e": "abc",
                        "f": "<class 'int'>",
                        "g": 0,
                    },
                    {
                        "a": None,
                        "b": False,
                        "c": 8,
                        "d": 9.1,
                        "e": "abc",
                        "f": "<class 'tests.util.test_serializable.PlainSimpleTypesContainer'>",
                        "g": 0,
                    },
                ],
                "u": None,
            },
            container.to_json(),
        )

    def test_dataclass_simple_ok(self):
        self.assertEqual(
            {},
            DataclassSimpleTypesContainer().to_json(),
        )
        self.assertEqual(
            {
                "a": "?",
                "b": True,
                "c": 12,
                "d": 34.56,
                "e": "uvw",
                "f": "<class 'bool'>",
                "g": "off",
            },
            DataclassSimpleTypesContainer(
                a="?", b=True, c=12, d=34.56, e="uvw", f=bool, g="off"
            ).to_json(),
        )

    def test_dataclass_complex_ok(self):
        container = DataclassComplexTypesContainer(
            q=dict(p=True, q=False),
            r=dict(
                u=DataclassSimpleTypesContainer(), v=DataclassSimpleTypesContainer()
            ),
            s=[1, 2, 3],
            t=[
                DataclassSimpleTypesContainer(c=5, d=6.7),
                DataclassSimpleTypesContainer(
                    c=8, d=9.1, f=DataclassSimpleTypesContainer
                ),
            ],
        )
        self.assertEqual(
            {
                "p": {},
                "q": {"p": True, "q": False},
                "r": {"u": {}, "v": {}},
                "s": [1, 2, 3],
                "t": [
                    {"c": 5, "d": 6.7},
                    {
                        "c": 8,
                        "d": 9.1,
                        "f": (
                            "<class 'tests.util.test_serializable"
                            ".DataclassSimpleTypesContainer'>"
                        ),
                    },
                ],
            },
            container.to_json(),
        )

    def test_fail(self):
        @dataclass()
        class Problematic(JsonSerializable):
            data: Any

        with pytest.raises(
            TypeError,
            match=(
                "problematic.data must be of type"
                " None|bool|int|float|str|dict|list|tuple, but got object"
            ),
        ):
            Problematic(data=object()).to_json(value_name="problematic")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import html
from collections.abc import Iterable

from xrlint.constants import SEVERITY_CODE_TO_COLOR, SEVERITY_CODE_TO_NAME
from xrlint.formatter import FormatterContext, FormatterOp
from xrlint.formatters import registry
from xrlint.result import Message, Result, get_rules_meta_for_results
from xrlint.util.formatting import format_problems
from xrlint.util.schema import schema


@registry.define_formatter(
    "html",
    version="1.0.0",
    schema=schema(
        "object",
        properties=dict(
            with_meta=schema("boolean", default=False),
        ),
    ),
)
class Html(FormatterOp):
    def __init__(self, with_meta: bool = False):
        self.with_meta = with_meta

    def format(
        self,
        context: FormatterContext,
        results: Iterable[Result],
    ) -> str:
        results = list(results)  # get them all

        lines = [
            "<div>",
            "<h3>Results</h3>",
        ]
        for result in results:
            lines.extend(format_result(result))
        lines.append("</div>")

        if self.with_meta:
            rules_meta = get_rules_meta_for_results(results)
            lines.append("<div>")
            lines.append("<h3>Rules</h3>")
            for rm in rules_meta.values():
                lines.append(
                    f"<p>Rule <strong>{rm.name}</strong>, version {rm.version}</p>"
                )
                if rm.description:
                    lines.append(f"<p>{rm.description}</p>")
                if rm.docs_url:
                    lines.append(
                        f'<p><a href="{rm.docs_url}">Rule documentation</a></p>'
                    )
            lines.append("</div>")

        return HtmlText("\n".join(lines))


def format_result(result: Result) -> list[str]:
    lines = ['<div style="padding-bottom: 5px">']
    escaped_path = _format_file_path(result.file_path)
    if not result.messages:
        lines.append(
            f"<p>{escaped_path} - "
            f'<span style="font-weight:bold;color:green">ok</span>'
            f"</p>"
        )
    else:
        lines.append(
            f"<p>{escaped_path} - "
            f"<span>{format_problems(result.error_count, result.warning_count)}</span>"
            f"</p>"
        )
        lines.append("<hr/>")
        table_data = []
        for m in result.messages:
            table_data.append(
                [
                    _format_node_path(m),
                    _format_severity(m),
                    _format_message(m),
                    _format_rule_id(m, result),
                ]
            )
        lines.extend(_format_result_data(table_data))
    lines.append("</div>")
    return lines


def _format_file_path(file_path: str) -> str:
    return f'<span style="font-family:monospace;font-weight:bold">{html.escape(file_path)}</span>'


def _format_node_path(m: Message) -> str:
    if not m.node_path:
        return ""
    return f'<span style="font-family:monospace;font-size:0.7em">{m.node_path}</span>'


def _format_message(m: Message) -> str:
    return m.message


def _format_rule_id(m: Message, r: Result) -> str:
    if not m.rule_id:
        return ""
    docs_url = r.get_docs_url_for_rule(m.rule_id)
    if docs_url:
        return f'<a href="{docs_url}">{m.rule_id}</a>'
    return m.rule_id


def _format_severity(m: Message) -> str:
    if not m.severity:
        return ""
    name = SEVERITY_CODE_TO_NAME.get(m.severity)
    color = SEVERITY_CODE_TO_COLOR.get(m.severity)
    return f'<span style="color:{color}">{name}</span>'


def _format_result_data(data: list[list[str]]) -> list[str]:
    lines = ["<table>"]
    for row in data:
        lines.append("  <tr>")
        lines.extend(f'    <td  style="text-align:left">{value}</td>' for value in row)
        lines.append("  </tr>")
    lines.append("</table>")
    return lines


class HtmlText(str):
    """Allow displaying `text` as HTML in Jupyter notebooks."""

    def _repr_html_(self: str) -> str:
        """Represent HTML text as HTML."""
        return self



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).


def export_config():
    import xrlint.plugins.core
    import xrlint.plugins.xcube

    core = xrlint.plugins.core.export_plugin()
    xcube = xrlint.plugins.xcube.export_plugin()
    return [
        {
            "plugins": {
                "xcube": xcube,
            }
        },
        *core.configs["recommended"],
        *xcube.configs["recommended"],
        {
            "rules": {
                "xcube/dataset-title": "error",
                "xcube/single-grid-mapping": "off",
            }
        },
    ]



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from tests.formatters.helpers import get_context
from xrlint.config import ConfigObject
from xrlint.formatters.simple import Simple
from xrlint.result import Message, Result


class SimpleTest(TestCase):
    errors_and_warnings = [
        Result(
            file_path="test1.nc",
            config_object=ConfigObject(),
            messages=[
                Message(message="what", rule_id="rule-1", severity=2),
                Message(message="is", fatal=True),
                Message(message="happening?", rule_id="rule-2", severity=1),
            ],
        )
    ]

    warnings_only = [
        Result(
            file_path="test2.nc",
            config_object=ConfigObject(),
            messages=[
                Message(message="what", rule_id="rule-1", severity=1),
                Message(message="happened?", rule_id="rule-2", severity=1),
            ],
        )
    ]

    def test_no_color(self):
        formatter = Simple(styled=False)
        text = formatter.format(
            context=get_context(),
            results=self.errors_and_warnings,
        )
        self.assert_output_1_ok(text)
        self.assertNotIn("\033]", text)

        formatter = Simple(styled=False)
        text = formatter.format(
            context=get_context(),
            results=self.warnings_only,
        )
        self.assert_output_2_ok(text)
        self.assertNotIn("\033]", text)

    def test_color(self):
        formatter = Simple(styled=True)
        text = formatter.format(
            context=get_context(),
            results=self.errors_and_warnings,
        )
        self.assert_output_1_ok(text)
        self.assertIn("\033]", text)

        formatter = Simple(styled=True)
        text = formatter.format(
            context=get_context(),
            results=self.warnings_only,
        )
        self.assert_output_2_ok(text)
        self.assertIn("\033]", text)

    def assert_output_1_ok(self, text):
        self.assertIsInstance(text, str)
        self.assertIn("test1.nc", text)
        self.assertIn("happening?", text)
        self.assertIn("error", text)
        self.assertIn("warn", text)
        self.assertIn("rule-1", text)
        self.assertIn("rule-2", text)

    def assert_output_2_ok(self, text):
        self.assertIsInstance(text, str)
        self.assertIn("test2.nc", text)
        self.assertIn("happened?", text)
        self.assertNotIn("error", text)
        self.assertIn("warn", text)
        self.assertIn("rule-1", text)
        self.assertIn("rule-2", text)
